import pymysql
from getpass import getpass
from .connection import conn
from .settings import config
from .utils import user_choice


def set_password(new_password=None, connection=None, update_config=None):   # pragma: no cover
    connection = conn() if connection is None else connection
    if new_password is None:
        new_password = getpass('New password: ')
        confirm_password = getpass('Confirm password: ')
        if new_password != confirm_password:
            print('Failed to confirm the password! Aborting password change.')
            return
    connection.query("SET PASSWORD = PASSWORD('%s')" % new_password)
    print('Password updated.')

    if update_config or (update_config is None and user_choice('Update local setting?') == 'yes'):
        config['database.password'] = new_password
        config.save_local(verbose=True)


def kill(restriction=None, connection=None, order_by=None):  # pragma: no cover
    """
    view and kill database connections.
    :param restriction: restriction to be applied to processlist
    :param connection: a datajoint.Connection object. Default calls datajoint.conn()
    :param order_by: order by a single attribute or the list of attributes. defaults to 'id'.

    Restrictions are specified as strings and can involve any of the attributes of
    information_schema.processlist: ID, USER, HOST, DB, COMMAND, TIME, STATE, INFO.

    Examples:
        dj.kill('HOST LIKE "%compute%"') lists only connections from hosts containing "compute".
        dj.kill('TIME > 600') lists only connections in their current state for more than 10 minutes
    """

    if connection is None:
        connection = conn()

    if order_by is not None and not isinstance(order_by, str):
        order_by = ','.join(order_by)

    query = 'SELECT * FROM information_schema.processlist WHERE id <> CONNECTION_ID()' + (
        "" if restriction is None else ' AND (%s)' % restriction) + (
            ' ORDER BY %s' % (order_by or 'id'))

    while True:
        print('  ID USER         HOST          STATE         TIME    INFO')
        print('+--+ +----------+ +-----------+ +-----------+ +-----+')
        cur = ({k.lower(): v for k, v in elem.items()}
               for elem in connection.query(query, as_dict=True))
        for process in cur:
            try:
                print('{id:>4d} {user:<12s} {host:<12s} {state:<12s} {time:>7d}  {info}'.format(**process))
            except TypeError:
                print(process)
        response = input('process to kill or "q" to quit > ')
        if response == 'q':
            break
        if response:
            try:
                pid = int(response)
            except ValueError:
                pass  # ignore non-numeric input
            else:
                try:
                    connection.query('kill %d' % pid)
                except pymysql.err.InternalError:
                    print('Process not found')


def kill_quick(restriction=None, connection=None):
    """
    Kill database connections without prompting. Returns number of terminated connections.
    :param restriction: restriction to be applied to processlist
    :param connection: a datajoint.Connection object. Default calls datajoint.conn()

    Restrictions are specified as strings and can involve any of the attributes of
    information_schema.processlist: ID, USER, HOST, DB, COMMAND, TIME, STATE, INFO.

    Examples:
        dj.kill('HOST LIKE "%compute%"') terminates connections from hosts containing "compute".
    """
    if connection is None:
        connection = conn()

    query = 'SELECT * FROM information_schema.processlist WHERE id <> CONNECTION_ID()' + (
        "" if restriction is None else ' AND (%s)' % restriction)

    cur = ({k.lower(): v for k, v in elem.items()}
           for elem in connection.query(query, as_dict=True))
    nkill = 0
    for process in cur:
        connection.query('kill %d' % process['id'])
        nkill += 1
    return nkill



from functools import partial
from pathlib import Path
import warnings
import pandas
import itertools
import re
import numpy as np
import uuid
import numbers
from . import blob, hash
from .errors import DataJointError
from .settings import config
from .utils import safe_write


class key:
    """
    object that allows requesting the primary key as an argument in expression.fetch()
    The string "KEY" can be used instead of the class key
    """
    pass


def _get_download_path(download_path):
    """function to get download_path
    """

    if download_path is None and config['tmp_folder'] is None:
        return '.'
    elif download_path is not None:
        return download_path
    else:
        return config['tmp_folder']


def is_key(attr):
    return attr is key or attr == 'KEY'


def to_dicts(recarray):
    """convert record array to a dictionaries"""
    for rec in recarray:
        yield dict(zip(recarray.dtype.names, rec.tolist()))


def _get(connection, attr, data, squeeze, download_path, apply_adapter, get_external):
    """
    This function is called for every attribute

    :param connection: a dj.Connection object
    :param attr: attribute name from the table's heading
    :param data: literal value fetched from the table
    :param squeeze: if True squeeze blobs
    :param download_path: for fetches that download data, e.g. attachments
    :param apply_adapter: whether to apply the adapter
    :return: unpacked data
    """
    if data is None:
        return

    # get correct download path
    download_path = _get_download_path(download_path)

    extern = connection.schemas[attr.database].external[attr.store] if attr.is_external else None

    # apply attribute adapter if present
    adapt = attr.adapter.get if attr.adapter and apply_adapter else lambda x: x

    if not get_external and attr.is_external:
        return uuid.UUID(bytes=data)

    if attr.is_filepath:
        return adapt(extern.download_filepath(uuid.UUID(bytes=data))[0])

    if attr.is_attachment:
        # Steps:
        # 1. get the attachment filename
        # 2. check if the file already exists at download_path, verify checksum
        # 3. if exists and checksum passes then return the local filepath
        # 4. Otherwise, download the remote file and return the new filepath
        _uuid = uuid.UUID(bytes=data) if attr.is_external else None
        attachment_name = (extern.get_attachment_name(_uuid) if attr.is_external
                           else data.split(b"\0", 1)[0].decode())
        local_filepath = Path(download_path) / attachment_name
        if local_filepath.is_file():
            attachment_checksum = _uuid if attr.is_external else hash.uuid_from_buffer(data)
            if attachment_checksum == hash.uuid_from_file(local_filepath, init_string=attachment_name + '\0'):
                return adapt(str(local_filepath))  # checksum passed, no need to download again
            # generate the next available alias filename
            for n in itertools.count():
                f = local_filepath.parent / (local_filepath.stem + '_%04x' % n + local_filepath.suffix)
                if not f.is_file():
                    local_filepath = f
                    break
                if attachment_checksum == hash.uuid_from_file(f, init_string=attachment_name + '\0'):
                    return adapt(str(f))  # checksum passed, no need to download again
        # Save attachment
        if attr.is_external:
            extern.download_attachment(_uuid, attachment_name, local_filepath)
        else:
            # write from buffer
            safe_write(local_filepath, data.split(b"\0", 1)[1])
        return adapt(str(local_filepath))  # download file from remote store

    return adapt(uuid.UUID(bytes=data) if attr.uuid else (
            blob.unpack(extern.get(uuid.UUID(bytes=data)) if attr.is_external else data, squeeze=squeeze)
            if attr.is_blob else data))


def _flatten_attribute_list(primary_key, attrs):
    """
    :param primary_key: list of attributes in primary key
    :param attrs: list of attribute names, which may include "KEY", "KEY DESC" or "KEY ASC"
    :return: generator of attributes where "KEY" is replaces with its component attributes
    """
    for a in attrs:
        if re.match(r'^\s*KEY(\s+[aA][Ss][Cc])?\s*$', a):
            yield from primary_key
        elif re.match(r'^\s*KEY\s+[Dd][Ee][Ss][Cc]\s*$', a):
            yield from (q + ' DESC' for q in primary_key)
        else:
            yield a


class Fetch:
    """
    A fetch object that handles retrieving elements from the table expression.
    :param expression: the QueryExpression object to fetch from.
    """

    def __init__(self, expression):
        self._expression = expression

    def __call__(self, *attrs, offset=None, limit=None, order_by=None, format=None, as_dict=None,
                 squeeze=False, download_path=None, apply_adapter=True, get_external=True):
        """
        Fetches the expression results from the database into an np.array or list of dictionaries and
        unpacks blob attributes.
        :param attrs: zero or more attributes to fetch. If not provided, the call will return
        all attributes of this relation. If provided, returns tuples with an entry for each attribute.
        :param offset: the number of tuples to skip in the returned result
        :param limit: the maximum number of tuples to return
        :param order_by: a single attribute or the list of attributes to order the results.
                No ordering should be assumed if order_by=None.
                To reverse the order, add DESC to the attribute name or names: e.g. ("age DESC", "frequency")
                To order by primary key, use "KEY" or "KEY DESC"
        :param format: Effective when as_dict=None and when attrs is empty
                None: default from config['fetch_format'] or 'array' if not configured
                "array": use numpy.key_array
                "frame": output pandas.DataFrame. .
        :param as_dict: returns a list of dictionaries instead of a record array.
                Defaults to False for .fetch() and to True for .fetch('KEY')
        :param squeeze:  if True, remove extra dimensions from arrays
        :param download_path: for fetches that download data, e.g. attachments
        :param apply_adapter: whether to apply the adapter
        :param get_external: whether to return the external blob or attachment or only the uuid
        :return: the contents of the relation in the form of a structured numpy.array or a dict list
        """
        if order_by is not None:
            # if 'order_by' passed in a string, make into list
            if isinstance(order_by, str):
                order_by = [order_by]
            # expand "KEY" or "KEY DESC"
            order_by = list(_flatten_attribute_list(self._expression.primary_key, order_by))

        attrs_as_dict = as_dict and attrs
        if attrs_as_dict:
            # absorb KEY into attrs and prepare to return attributes as dict (issue #595)
            if any(is_key(k) for k in attrs):
                attrs = list(self._expression.primary_key) + [
                    a for a in attrs if a not in self._expression.primary_key]
        if as_dict is None:
            as_dict = bool(attrs)  # default to True for "KEY" and False otherwise
        # format should not be specified with attrs or is_dict=True
        if format is not None and (as_dict or attrs):
            raise DataJointError('Cannot specify output format when as_dict=True or '
                                 'when attributes are selected to be fetched separately.')
        if format not in {None, "array", "frame"}:
            raise DataJointError(
                'Fetch output format must be in '
                '{{"array", "frame"}} but "{}" was given'.format(format))

        if not (attrs or as_dict) and format is None:
            format = config['fetch_format']  # default to array
            if format not in {"array", "frame"}:
                raise DataJointError(
                    'Invalid entry "{}" in datajoint.config["fetch_format"]: '
                    'use "array" or "frame"'.format(format))

        if limit is None and offset is not None:
            warnings.warn('Offset set, but no limit. Setting limit to a large number. '
                          'Consider setting a limit explicitly.')
            limit = 8000000000  # just a very large number to effect no limit

        get = partial(_get, self._expression.connection,
                      apply_adapter=apply_adapter, get_external=get_external,
                      squeeze=squeeze, download_path=download_path)
        if attrs:  # a list of attributes provided
            attributes = [a for a in attrs if not is_key(a)]
            ret = self._expression.proj(*attributes)
            ret = ret.fetch(
                offset=offset, limit=limit, order_by=order_by,
                as_dict=False, squeeze=squeeze, download_path=download_path,
                format='array')
            if attrs_as_dict:
                ret = [{k: v for k, v in zip(ret.dtype.names, x) if k in attrs} for x in ret]
            else:
                return_values = [list(
                    (to_dicts if as_dict else lambda x: x)(ret[self._expression.primary_key]))
                                 if is_key(attribute) else ret[attribute]
                                 for attribute in attrs]
                ret = return_values[0] if len(attrs) == 1 else return_values
        else:  # fetch all attributes as a numpy.record_array or pandas.DataFrame
            cur = self._expression.cursor(
                as_dict=as_dict, limit=limit, offset=offset, order_by=order_by)
            heading = self._expression.heading
            if as_dict:
                ret = [dict((name, get(heading[name], d[name]))
                            for name in heading.names) for d in cur]
            else:
                ret = list(cur.fetchall())
                record_type = (heading.as_dtype if not ret else np.dtype(
                    [(name, type(value))   # use the first element to determine blob type
                        if heading[name].is_blob and isinstance(value, numbers.Number)
                        else (name, heading.as_dtype[name])
                        for value, name in zip(ret[0], heading.as_dtype.names)]))
                try:
                    ret = np.array(ret, dtype=record_type)
                except Exception as e:
                    raise e
                for name in heading:
                    # unpack blobs and externals
                    ret[name] = list(map(partial(get, heading[name]), ret[name]))
                if format == "frame":
                    ret = pandas.DataFrame(ret).set_index(heading.primary_key)
        return ret


class Fetch1:
    """
    Fetch object for fetching the result of a query yielding one row.
    :param expression: a query expression to fetch from.
    """
    def __init__(self, expression):
        self._expression = expression

    def __call__(self, *attrs, squeeze=False, download_path=None,
                 apply_adapter=True, get_external=True):
        """
        Fetches the result of a query expression that yields one entry.

        If no attributes are specified, returns the result as a dict.
        If attributes are specified returns the corresponding results as a tuple.

        Examples:
        d = rel.fetch1()   # as a dictionary
        a, b = rel.fetch1('a', 'b')   # as a tuple

        :params *attrs: attributes to return when expanding into a tuple.
                 If attrs is empty, the return result is a dict
        :param squeeze:  When true, remove extra dimensions from arrays in attributes
        :param download_path: for fetches that download data, e.g. attachments
        :param apply_adapter: whether to apply the adapter
        :param get_external: whether to return the external blob or attachment or only the uuid
        :return: the one tuple in the relation in the form of a dict
        """
        heading = self._expression.heading

        if not attrs:  # fetch all attributes, return as ordered dict
            cur = self._expression.cursor(as_dict=True)
            ret = cur.fetchone()
            if not ret or cur.fetchone():
                raise DataJointError('fetch1 requires exactly one tuple in the input set.')
            ret = dict((name, _get(self._expression.connection, heading[name], ret[name],
                                   squeeze=squeeze, download_path=download_path,
                                   apply_adapter=apply_adapter, get_external=get_external))
                       for name in heading.names)
        else:  # fetch some attributes, return as tuple
            attributes = [a for a in attrs if not is_key(a)]
            result = self._expression.proj(*attributes).fetch(
                squeeze=squeeze, download_path=download_path, format="array")
            if len(result) != 1:
                raise DataJointError(
                    'fetch1 should only return one tuple. %d tuples found' % len(result))
            return_values = tuple(
                next(to_dicts(result[self._expression.primary_key]))
                if is_key(attribute) else result[attribute][0]
                for attribute in attrs)
            ret = return_values[0] if len(attrs) == 1 else return_values
        return ret



""" methods for generating SQL WHERE clauses from datajoint restriction conditions """

import inspect
import collections
import re
import uuid
import datetime
import decimal
import numpy
import pandas
from .errors import DataJointError


class PromiscuousOperand:
    """
    A container for an operand to ignore join compatibility
    """
    def __init__(self, operand):
        self.operand = operand


class AndList(list):
    """
    A list of conditions to by applied to a query expression by logical conjunction: the conditions are AND-ed.
    All other collections (lists, sets, other entity sets, etc) are applied by logical disjunction (OR).

    Example:
    expr2 = expr & dj.AndList((cond1, cond2, cond3))
    is equivalent to
    expr2 = expr & cond1 & cond2 & cond3
    """
    def append(self, restriction):
        if isinstance(restriction, AndList):
            # extend to reduce nesting
            self.extend(restriction)
        else:
            super().append(restriction)


class Not:
    """ invert restriction """
    def __init__(self, restriction):
        self.restriction = restriction


def assert_join_compatibility(expr1, expr2):
    """
    Determine if expressions expr1 and expr2 are join-compatible.  To be join-compatible,
    the matching attributes in the two expressions must be in the primary key of one or the
    other expression.
    Raises an exception if not compatible.
    :param expr1: A QueryExpression object
    :param expr2: A QueryExpression object
    """
    from .expression import QueryExpression, U

    for rel in (expr1, expr2):
        if not isinstance(rel, (U, QueryExpression)):
            raise DataJointError('Object %r is not a QueryExpression and cannot be joined.' % rel)
    if not isinstance(expr1, U) and not isinstance(expr2, U):  # dj.U is always compatible
        try:
            raise DataJointError(
                "Cannot join query expressions on dependent attribute `%s`" % next(
                    r for r in set(expr1.heading.secondary_attributes).intersection(
                        expr2.heading.secondary_attributes)))
        except StopIteration:
            pass  # all ok


def make_condition(query_expression, condition, columns):
    """
    Translate the input condition into the equivalent SQL condition (a string)
    :param query_expression: a dj.QueryExpression object to apply condition
    :param condition: any valid restriction object.
    :param columns: a set passed by reference to collect all column names used in the condition.
    :return: an SQL condition string or a boolean value.
    """
    from .expression import QueryExpression, Aggregation, U

    def prep_value(k, v):
        """prepare value v for inclusion as a string in an SQL condition"""
        if query_expression.heading[k].uuid:
            if not isinstance(v, uuid.UUID):
                try:
                    v = uuid.UUID(v)
                except (AttributeError, ValueError):
                    raise DataJointError(
                        'Badly formed UUID {v} in restriction by `{k}`'.format(k=k, v=v))
            return "X'%s'" % v.bytes.hex()
        if isinstance(v, (datetime.date, datetime.datetime, datetime.time, decimal.Decimal)):
            return '"%s"' % v
        if isinstance(v, str):
            return '"%s"' % v.replace('%', '%%')
        return '%r' % v

    negate = False
    while isinstance(condition, Not):
        negate = not negate
        condition = condition.restriction
    template = "NOT (%s)" if negate else "%s"

    # restrict by string
    if isinstance(condition, str):
        columns.update(extract_column_names(condition))
        return template % condition.strip().replace("%", "%%")  # escape % in strings, see issue #376

    # restrict by AndList
    if isinstance(condition, AndList):
        # omit all conditions that evaluate to True
        items = [item for item in (make_condition(query_expression, cond, columns) for cond in condition)
                 if item is not True]
        if any(item is False for item in items):
            return negate  # if any item is False, the whole thing is False
        if not items:
            return not negate  # and empty AndList is True
        return template % ('(' + ') AND ('.join(items) + ')')

    # restriction by dj.U evaluates to True
    if isinstance(condition, U):
        return not negate

    # restrict by boolean
    if isinstance(condition, bool):
        return negate != condition

    # restrict by a mapping such as a dict -- convert to an AndList of string equality conditions
    if isinstance(condition, collections.abc.Mapping):
        common_attributes = set(condition).intersection(query_expression.heading.names)
        if not common_attributes:
            return not negate  # no matching attributes -> evaluates to True
        columns.update(common_attributes)
        return template % ('(' + ') AND ('.join(
            '`%s`=%s' % (k, prep_value(k, condition[k])) for k in common_attributes) + ')')

    # restrict by a numpy record -- convert to an AndList of string equality conditions
    if isinstance(condition, numpy.void):
        common_attributes = set(condition.dtype.fields).intersection(query_expression.heading.names)
        if not common_attributes:
            return not negate   # no matching attributes -> evaluate to True
        columns.update(common_attributes)
        return template % ('(' + ') AND ('.join(
            '`%s`=%s' % (k, prep_value(k, condition[k])) for k in common_attributes) + ')')

    # restrict by a QueryExpression subclass -- trigger instantiation and move on
    if inspect.isclass(condition) and issubclass(condition, QueryExpression):
        condition = condition()

    # restrict by another expression (aka semijoin and antijoin)
    check_compatibility = True
    if isinstance(condition, PromiscuousOperand):
        condition = condition.operand
        check_compatibility = False

    if isinstance(condition, QueryExpression):
        if check_compatibility:
            assert_join_compatibility(query_expression, condition)
        common_attributes = [q for q in condition.heading.names if q in query_expression.heading.names]
        columns.update(common_attributes)
        if isinstance(condition, Aggregation):
            condition = condition.make_subquery()
        return (
            # without common attributes, any non-empty set matches everything
            (not negate if condition else negate) if not common_attributes
            else '({fields}) {not_}in ({subquery})'.format(
                fields='`' + '`,`'.join(common_attributes) + '`',
                not_="not " if negate else "",
                subquery=condition.make_sql(common_attributes)))

    # restrict by pandas.DataFrames
    if isinstance(condition, pandas.DataFrame):
        condition = condition.to_records()  # convert to numpy.recarray and move on

    # if iterable (but not a string, a QueryExpression, or an AndList), treat as an OrList
    try:
        or_list = [make_condition(query_expression, q, columns) for q in condition]
    except TypeError:
        raise DataJointError('Invalid restriction type %r' % condition)
    else:
        or_list = [item for item in or_list if item is not False]  # ignore all False conditions
        if any(item is True for item in or_list):  # if any item is True, the whole thing is True
            return not negate
        return template % ('(%s)' % ' OR '.join(or_list)) if or_list else negate  # an empty or list is False


def extract_column_names(sql_expression):
    """
    extract all presumed column names from an sql expression such as the WHERE clause, for example.
    :param sql_expression: a string containing an SQL expression
    :return: set of extracted column names
    This may be MySQL-specific for now.
    """
    assert isinstance(sql_expression, str)
    result = set()
    s = sql_expression  # for terseness
    # remove escaped quotes
    s = re.sub(r'(\\\")|(\\\')', '', s)
    # remove quoted text
    s = re.sub(r"'[^']*'", "", s)
    s = re.sub(r'"[^"]*"', '', s)
    # find all tokens in back quotes and remove them
    result.update(re.findall(r"`([a-z][a-z_0-9]*)`", s))
    s = re.sub(r"`[a-z][a-z_0-9]*`", '', s)
    # remove space before parentheses
    s = re.sub(r"\s*\(", "(", s)
    # remove tokens followed by ( since they must be functions
    s = re.sub(r"(\b[a-z][a-z_0-9]*)\(", "(", s)
    remaining_tokens = set(re.findall(r"\b[a-z][a-z_0-9]*\b", s))
    # update result removing reserved words
    result.update(remaining_tokens - {"is", "in", "between", "like", "and", "or", "null", "not"})
    return result



from pathlib import Path, PurePosixPath, PureWindowsPath
from collections import Mapping
from tqdm import tqdm
from .settings import config
from .errors import DataJointError, MissingExternalFile
from .hash import uuid_from_buffer, uuid_from_file
from .table import Table, FreeTable
from .heading import Heading
from .declare import EXTERNAL_TABLE_ROOT
from . import s3
from .utils import safe_write, safe_copy

CACHE_SUBFOLDING = (2, 2)   # (2, 2) means  "0123456789abcd" will be saved as "01/23/0123456789abcd"
SUPPORT_MIGRATED_BLOBS = True   # support blobs migrated from datajoint 0.11.*


def subfold(name, folds):
    """
    subfolding for external storage:   e.g.  subfold('aBCdefg', (2, 3))  -->  ['ab','cde']
    """
    return (name[:folds[0]].lower(),) + subfold(name[folds[0]:], folds[1:]) if folds else ()


class ExternalTable(Table):
    """
    The table tracking externally stored objects.
    Declare as ExternalTable(connection, database)
    """
    def __init__(self, connection, store, database):
        self.store = store
        self.spec = config.get_store_spec(store)
        self._s3 = None
        self.database = database
        self._connection = connection
        self._heading = Heading(table_info=dict(
            conn=connection,
            database=database,
            table_name=self.table_name,
            context=None))
        self._support = [self.full_table_name]
        if not self.is_declared:
            self.declare()
        self._s3 = None
        if self.spec['protocol'] == 'file' and not Path(self.spec['location']).is_dir():
            raise FileNotFoundError('Inaccessible local directory %s' %
                                    self.spec['location']) from None

    @property
    def definition(self):
        return """
        # external storage tracking
        hash  : uuid    #  hash of contents (blob), of filename + contents (attach), or relative filepath (filepath)
        ---
        size      :bigint unsigned     # size of object in bytes
        attachment_name=null : varchar(255)  # the filename of an attachment
        filepath=null : varchar(1000)  # relative filepath or attachment filename
        contents_hash=null : uuid      # used for the filepath datatype
        timestamp=CURRENT_TIMESTAMP  :timestamp   # automatic timestamp
        """

    @property
    def table_name(self):
        return '{external_table_root}_{store}'.format(external_table_root=EXTERNAL_TABLE_ROOT, store=self.store)

    @property
    def s3(self):
        if self._s3 is None:
            self._s3 = s3.Folder(**self.spec)
        return self._s3

    # - low-level operations - private

    def _make_external_filepath(self, relative_filepath):
        """resolve the complete external path based on the relative path"""
        # Strip root
        if self.spec['protocol'] == 's3':
            posix_path = PurePosixPath(PureWindowsPath(self.spec['location']))
            location_path = Path(
                    *posix_path.parts[1:]) if len(
                    self.spec['location']) > 0 and any(
                    case in posix_path.parts[0] for case in (
                        '\\', ':')) else Path(posix_path)
            return PurePosixPath(location_path, relative_filepath)
        # Preserve root
        elif self.spec['protocol'] == 'file':
            return PurePosixPath(Path(self.spec['location']), relative_filepath)
        else:
            assert False

    def _make_uuid_path(self, uuid, suffix=''):
        """create external path based on the uuid hash"""
        return self._make_external_filepath(PurePosixPath(
            self.database, '/'.join(subfold(uuid.hex, self.spec['subfolding'])), uuid.hex).with_suffix(suffix))

    def _upload_file(self, local_path, external_path, metadata=None):
        if self.spec['protocol'] == 's3':
            self.s3.fput(local_path, external_path, metadata)
        elif self.spec['protocol'] == 'file':
            safe_copy(local_path, external_path, overwrite=True)
        else:
            assert False

    def _download_file(self, external_path, download_path):
        if self.spec['protocol'] == 's3':
            self.s3.fget(external_path, download_path)
        elif self.spec['protocol'] == 'file':
            safe_copy(external_path, download_path)
        else:
            assert False

    def _upload_buffer(self, buffer, external_path):
        if self.spec['protocol'] == 's3':
            self.s3.put(external_path, buffer)
        elif self.spec['protocol'] == 'file':
            safe_write(external_path, buffer)
        else:
            assert False

    def _download_buffer(self, external_path):
        if self.spec['protocol'] == 's3':
            return self.s3.get(external_path)
        if self.spec['protocol'] == 'file':
            return Path(external_path).read_bytes()
        assert False

    def _remove_external_file(self, external_path):
        if self.spec['protocol'] == 's3':
            self.s3.remove_object(external_path)
        elif self.spec['protocol'] == 'file':
            Path(external_path).unlink()

    def exists(self, external_filepath):
        """
        :return: True if the external file is accessible
        """
        if self.spec['protocol'] == 's3':
            return self.s3.exists(external_filepath)
        if self.spec['protocol'] == 'file':
            return Path(external_filepath).is_file()
        assert False

    # --- BLOBS ----

    def put(self, blob):
        """
        put a binary string (blob) in external store
        """
        uuid = uuid_from_buffer(blob)
        self._upload_buffer(blob, self._make_uuid_path(uuid))
        # insert tracking info
        self.connection.query(
            "INSERT INTO {tab} (hash, size) VALUES (%s, {size}) ON DUPLICATE KEY "
            "UPDATE timestamp=CURRENT_TIMESTAMP".format(
                tab=self.full_table_name, size=len(blob)), args=(uuid.bytes,))
        return uuid

    def get(self, uuid):
        """
        get an object from external store.
        """
        if uuid is None:
            return None
        # attempt to get object from cache
        blob = None
        cache_folder = config.get('cache', None)
        if cache_folder:
            try:
                cache_path = Path(cache_folder, *subfold(uuid.hex, CACHE_SUBFOLDING))
                cache_file = Path(cache_path, uuid.hex)
                blob = cache_file.read_bytes()
            except FileNotFoundError:
                pass  # not cached
        # download blob from external store
        if blob is None:
            try:
                blob = self._download_buffer(self._make_uuid_path(uuid))
            except MissingExternalFile:
                if not SUPPORT_MIGRATED_BLOBS:
                    raise
                # blobs migrated from datajoint 0.11 are stored at explicitly defined filepaths
                relative_filepath, contents_hash = (self & {'hash': uuid}).fetch1('filepath', 'contents_hash')
                if relative_filepath is None:
                    raise
                blob = self._download_buffer(self._make_external_filepath(relative_filepath))
            if cache_folder:
                cache_path.mkdir(parents=True, exist_ok=True)
                safe_write(cache_path / uuid.hex, blob)
        return blob

    # --- ATTACHMENTS ---

    def upload_attachment(self, local_path):
        attachment_name = Path(local_path).name
        uuid = uuid_from_file(local_path, init_string=attachment_name + '\0')
        external_path = self._make_uuid_path(uuid, '.' + attachment_name)
        self._upload_file(local_path, external_path)
        # insert tracking info
        self.connection.query("""
        INSERT INTO {tab} (hash, size, attachment_name)
        VALUES (%s, {size}, "{attachment_name}")
        ON DUPLICATE KEY UPDATE timestamp=CURRENT_TIMESTAMP""".format(
                tab=self.full_table_name,
                size=Path(local_path).stat().st_size,
                attachment_name=attachment_name), args=[uuid.bytes])
        return uuid

    def get_attachment_name(self, uuid):
        return (self & {'hash': uuid}).fetch1('attachment_name')

    def download_attachment(self, uuid, attachment_name, download_path):
        """ save attachment from memory buffer into the save_path """
        external_path = self._make_uuid_path(uuid, '.' + attachment_name)
        self._download_file(external_path, download_path)

    # --- FILEPATH ---

    def upload_filepath(self, local_filepath):
        """
        Raise exception if an external entry already exists with a different contents checksum.
        Otherwise, copy (with overwrite) file to remote and
        If an external entry exists with the same checksum, then no copying should occur
        """
        local_filepath = Path(local_filepath)
        try:
            relative_filepath = str(local_filepath.relative_to(self.spec['stage']).as_posix())
        except ValueError:
            raise DataJointError('The path {path} is not in stage {stage}'.format(
                path=local_filepath.parent, **self.spec))
        uuid = uuid_from_buffer(init_string=relative_filepath)  # hash relative path, not contents
        contents_hash = uuid_from_file(local_filepath)

        # check if the remote file already exists and verify that it matches
        check_hash = (self & {'hash': uuid}).fetch('contents_hash')
        if check_hash:
            # the tracking entry exists, check that it's the same file as before
            if contents_hash != check_hash[0]:
                raise DataJointError(
                    "A different version of '{file}' has already been placed.".format(file=relative_filepath))
        else:
            # upload the file and create its tracking entry
            self._upload_file(local_filepath, self._make_external_filepath(relative_filepath),
                              metadata={'contents_hash': str(contents_hash)})
            self.connection.query(
                "INSERT INTO {tab} (hash, size, filepath, contents_hash) VALUES (%s, {size}, '{filepath}', %s)".format(
                    tab=self.full_table_name, size=Path(local_filepath).stat().st_size,
                    filepath=relative_filepath), args=(uuid.bytes, contents_hash.bytes))
        return uuid

    def download_filepath(self, filepath_hash):
        """
        sync a file from external store to the local stage
        :param filepath_hash: The hash (UUID) of the relative_path
        :return: hash (UUID) of the contents of the downloaded file or Nones
        """
        if filepath_hash is not None:
            relative_filepath, contents_hash = (self & {'hash': filepath_hash}).fetch1('filepath', 'contents_hash')
            external_path = self._make_external_filepath(relative_filepath)
            local_filepath = Path(self.spec['stage']).absolute() / relative_filepath
            file_exists = Path(local_filepath).is_file() and uuid_from_file(local_filepath) == contents_hash
            if not file_exists:
                self._download_file(external_path, local_filepath)
                checksum = uuid_from_file(local_filepath)
                if checksum != contents_hash:  # this should never happen without outside interference
                    raise DataJointError("'{file}' downloaded but did not pass checksum'".format(file=local_filepath))
            return str(local_filepath), contents_hash

    # --- UTILITIES ---

    @property
    def references(self):
        """
        :return: generator of referencing table names and their referencing columns
        """
        return ({k.lower(): v for k, v in elem.items()} for elem in self.connection.query("""
        SELECT concat('`', table_schema, '`.`', table_name, '`') as referencing_table, column_name
        FROM information_schema.key_column_usage
        WHERE referenced_table_name="{tab}" and referenced_table_schema="{db}"
        """.format(tab=self.table_name, db=self.database), as_dict=True))

    def fetch_external_paths(self, **fetch_kwargs):
        """
        generate complete external filepaths from the query.
        Each element is a tuple: (uuid, path)
        :param fetch_kwargs: keyword arguments to pass to fetch
        """
        fetch_kwargs.update(as_dict=True)
        paths = []
        for item in self.fetch('hash', 'attachment_name', 'filepath', **fetch_kwargs):
            if item['attachment_name']:
                # attachments
                path = self._make_uuid_path(item['hash'], '.' + item['attachment_name'])
            elif item['filepath']:
                # external filepaths
                path = self._make_external_filepath(item['filepath'])
            else:
                # blobs
                path = self._make_uuid_path(item['hash'])
            paths.append((item['hash'], path))
        return paths

    def unused(self):
        """
        query expression for unused hashes
        :return: self restricted to elements that are not in use by any tables in the schema
        """
        return self - [FreeTable(self.connection, ref['referencing_table']).proj(hash=ref['column_name'])
                       for ref in self.references]

    def used(self):
        """
        query expression for used hashes
        :return: self restricted to elements that in use by tables in the schema
        """
        return self & [FreeTable(self.connection, ref['referencing_table']).proj(hash=ref['column_name'])
                       for ref in self.references]

    def delete(self, *, delete_external_files=None, limit=None, display_progress=True):
        """
        :param delete_external_files: True or False. If False, only the tracking info is removed from the
        external store table but the external files remain intact. If True, then the external files
        themselves are deleted too.
        :param limit: (integer) limit the number of items to delete
        :param display_progress: if True, display progress as files are cleaned up
        :return: yields
        """
        if delete_external_files not in (True, False):
            raise DataJointError("The delete_external_files argument must be set to either True or False in delete()")

        if not delete_external_files:
            self.unused().delete_quick()
        else:
            items = self.unused().fetch_external_paths(limit=limit)
            if display_progress:
                items = tqdm(items)
            # delete items one by one, close to transaction-safe
            error_list = []
            for uuid, external_path in items:
                try:
                    count = (self & {'hash': uuid}).delete_quick(get_count=True)  # optimize
                except Exception:
                    pass   # if delete failed, do not remove the external file
                else:
                    assert count in (0, 1)
                    try:
                        self._remove_external_file(external_path)
                    except Exception as error:
                        error_list.append((uuid, external_path, str(error)))
            return error_list


class ExternalMapping(Mapping):
    """
    The external manager contains all the tables for all external stores for a given schema
    :Example:
        e = ExternalMapping(schema)
        external_table = e[store]
    """
    def __init__(self, schema):
        self.schema = schema
        self._tables = {}

    def __repr__(self):
        return ("External file tables for schema `{schema}`:\n    ".format(schema=self.schema.database)
                + "\n    ".join('"{store}" {protocol}:{location}'.format(
                    store=k, **v.spec) for k, v in self.items()))

    def __getitem__(self, store):
        """
        Triggers the creation of an external table.
        Should only be used when ready to save or read from external storage.
        :param store: the name of the store
        :return: the ExternalTable object for the store
        """
        if store not in self._tables:
            self._tables[store] = ExternalTable(
                connection=self.schema.connection, store=store, database=self.schema.database)
        return self._tables[store]

    def __len__(self):
        return len(self._tables)

    def __iter__(self):
        return iter(self._tables)



"""
This module hosts functions to convert DataJoint table definitions into mysql table definitions, and to
declare the corresponding mysql tables.
"""
import re
import pyparsing as pp
import logging
import warnings
from .errors import DataJointError, _support_filepath_types, FILEPATH_FEATURE_SWITCH
from .attribute_adapter import get_adapter
from .settings import config

UUID_DATA_TYPE = 'binary(16)'
MAX_TABLE_NAME_LENGTH = 64
CONSTANT_LITERALS = {'CURRENT_TIMESTAMP'}  # SQL literals to be used without quotes (case insensitive)
EXTERNAL_TABLE_ROOT = '~external'

TYPE_PATTERN = {k: re.compile(v, re.I) for k, v in dict(
    INTEGER=r'((tiny|small|medium|big|)int|integer)(\s*\(.+\))?(\s+unsigned)?(\s+auto_increment)?|serial$',
    DECIMAL=r'(decimal|numeric)(\s*\(.+\))?(\s+unsigned)?$',
    FLOAT=r'(double|float|real)(\s*\(.+\))?(\s+unsigned)?$',
    STRING=r'(var)?char\s*\(.+\)$',
    ENUM=r'enum\s*\(.+\)$',
    BOOL=r'bool(ean)?$',   # aliased to tinyint(1)
    TEMPORAL=r'(date|datetime|time|timestamp|year)(\s*\(.+\))?$',
    INTERNAL_BLOB=r'(tiny|small|medium|long|)blob$',
    EXTERNAL_BLOB=r'blob@(?P<store>[a-z]\w*)$',
    INTERNAL_ATTACH=r'attach$',
    EXTERNAL_ATTACH=r'attach@(?P<store>[a-z]\w*)$',
    FILEPATH=r'filepath@(?P<store>[a-z]\w*)$',
    UUID=r'uuid$',
    ADAPTED=r'<.+>$'
).items()}

# custom types are stored in attribute comment
SPECIAL_TYPES = {'UUID', 'INTERNAL_ATTACH', 'EXTERNAL_ATTACH', 'EXTERNAL_BLOB', 'FILEPATH', 'ADAPTED'}
NATIVE_TYPES = set(TYPE_PATTERN) - SPECIAL_TYPES
EXTERNAL_TYPES = {'EXTERNAL_ATTACH', 'EXTERNAL_BLOB', 'FILEPATH'}  # data referenced by a UUID in external tables
SERIALIZED_TYPES = {'EXTERNAL_ATTACH', 'INTERNAL_ATTACH', 'EXTERNAL_BLOB', 'INTERNAL_BLOB'}  # requires packing data

assert set().union(SPECIAL_TYPES, EXTERNAL_TYPES, SERIALIZED_TYPES) <= set(TYPE_PATTERN)


def match_type(attribute_type):
    try:
        return next(category for category, pattern in TYPE_PATTERN.items() if pattern.match(attribute_type))
    except StopIteration:
        raise DataJointError("Unsupported attribute type {type}".format(type=attribute_type))


logger = logging.getLogger(__name__)


def build_foreign_key_parser_old():
    # old-style foreign key parser. Superseded by expression-based syntax. See issue #436
    # This will be deprecated in a future release.
    left = pp.Literal('(').suppress()
    right = pp.Literal(')').suppress()
    attribute_name = pp.Word(pp.srange('[a-z]'), pp.srange('[a-z0-9_]'))
    new_attrs = pp.Optional(left + pp.delimitedList(attribute_name) + right).setResultsName('new_attrs')
    arrow = pp.Literal('->').suppress()
    lbracket = pp.Literal('[').suppress()
    rbracket = pp.Literal(']').suppress()
    option = pp.Word(pp.srange('[a-zA-Z]'))
    options = pp.Optional(lbracket + pp.delimitedList(option) + rbracket).setResultsName('options')
    ref_table = pp.Word(pp.alphas, pp.alphanums + '._').setResultsName('ref_table')
    ref_attrs = pp.Optional(left + pp.delimitedList(attribute_name) + right).setResultsName('ref_attrs')
    return new_attrs + arrow + options + ref_table + ref_attrs


def build_foreign_key_parser():
    arrow = pp.Literal('->').suppress()
    lbracket = pp.Literal('[').suppress()
    rbracket = pp.Literal(']').suppress()
    option = pp.Word(pp.srange('[a-zA-Z]'))
    options = pp.Optional(lbracket + pp.delimitedList(option) + rbracket).setResultsName('options')
    ref_table = pp.restOfLine.setResultsName('ref_table')
    return arrow + options + ref_table


def build_attribute_parser():
    quoted = pp.QuotedString('"') ^ pp.QuotedString("'")
    colon = pp.Literal(':').suppress()
    attribute_name = pp.Word(pp.srange('[a-z]'), pp.srange('[a-z0-9_]')).setResultsName('name')
    data_type = (pp.Combine(pp.Word(pp.alphas) + pp.SkipTo("#", ignore=quoted))
                 ^ pp.QuotedString('<', endQuoteChar='>', unquoteResults=False)).setResultsName('type')
    default = pp.Literal('=').suppress() + pp.SkipTo(colon, ignore=quoted).setResultsName('default')
    comment = pp.Literal('#').suppress() + pp.restOfLine.setResultsName('comment')
    return attribute_name + pp.Optional(default) + colon + data_type + comment


def build_index_parser():
    left = pp.Literal('(').suppress()
    right = pp.Literal(')').suppress()
    unique = pp.Optional(pp.CaselessKeyword('unique')).setResultsName('unique')
    index = pp.CaselessKeyword('index').suppress()
    attribute_name = pp.Word(pp.srange('[a-z]'), pp.srange('[a-z0-9_]'))
    return unique + index + left + pp.delimitedList(attribute_name).setResultsName('attr_list') + right


foreign_key_parser_old = build_foreign_key_parser_old()
foreign_key_parser = build_foreign_key_parser()
attribute_parser = build_attribute_parser()
index_parser = build_index_parser()


def is_foreign_key(line):
    """
    :param line: a line from the table definition
    :return: true if the line appears to be a foreign key definition
    """
    arrow_position = line.find('->')
    return arrow_position >= 0 and not any(c in line[:arrow_position] for c in '"#\'')


def compile_foreign_key(line, context, attributes, primary_key, attr_sql, foreign_key_sql, index_sql):
    """
    :param line: a line from a table definition
    :param context: namespace containing referenced objects
    :param attributes: list of attribute names already in the declaration -- to be updated by this function
    :param primary_key: None if the current foreign key is made from the dependent section. Otherwise it is the list
        of primary key attributes thus far -- to be updated by the function
    :param attr_sql: list of sql statements defining attributes -- to be updated by this function.
    :param foreign_key_sql: list of sql statements specifying foreign key constraints -- to be updated by this function.
    :param index_sql: list of INDEX declaration statements, duplicate or redundant indexes are ok.
    """
    # Parse and validate
    from .table import Table
    from .expression import QueryExpression

    obsolete = False   # See issue #436.  Old style to be deprecated in a future release
    try:
        result = foreign_key_parser.parseString(line)
    except pp.ParseException:
        try:
            result = foreign_key_parser_old.parseString(line)
        except pp.ParseBaseException as err:
            raise DataJointError('Parsing error in line "%s". %s.' % (line, err))
        else:
            obsolete = True
    try:
        ref = config.eval_from_context(context, result.ref_table)
    except NameError if obsolete else Exception:
        raise DataJointError('Foreign key reference %s could not be resolved' % result.ref_table)

    options = [opt.upper() for opt in result.options]
    for opt in options:  # check for invalid options
        if opt not in {'NULLABLE', 'UNIQUE'}:
            raise DataJointError('Invalid foreign key option "{opt}"'.format(opt=opt))
    is_nullable = 'NULLABLE' in options
    is_unique = 'UNIQUE' in options
    if is_nullable and primary_key is not None:
        raise DataJointError('Primary dependencies cannot be nullable in line "{line}"'.format(line=line))

    if obsolete:
        warnings.warn(
            'Line "{line}" uses obsolete syntax that will no longer be supported in datajoint 0.14. '
            'For details, see issue #780 https://github.com/datajoint/datajoint-python/issues/780'.format(line=line))
        if not isinstance(ref, type) or not issubclass(ref, Table):
            raise DataJointError('Foreign key reference %r must be a valid query' % result.ref_table)

    if isinstance(ref, type) and issubclass(ref, Table):
        ref = ref()

    # check that dependency is of a supported type
    if (not isinstance(ref, QueryExpression) or len(ref.restriction) or
            len(ref.support) != 1 or not isinstance(ref.support[0], str)):
        raise DataJointError('Dependency "%s" is not supported (yet). Use a base table or its projection.' %
                             result.ref_table)

    if obsolete:
        # for backward compatibility with old-style dependency declarations.  See issue #436
        if not isinstance(ref, Table):
            DataJointError('Dependency "%s" is not supported. Check documentation.' % result.ref_table)
        if not all(r in ref.primary_key for r in result.ref_attrs):
            raise DataJointError('Invalid foreign key attributes in "%s"' % line)
        try:
            raise DataJointError('Duplicate attributes "{attr}" in "{line}"'.format(
                attr=next(attr for attr in result.new_attrs if attr in attributes), line=line))
        except StopIteration:
            pass   # the normal outcome

        # Match the primary attributes of the referenced table to local attributes
        new_attrs = list(result.new_attrs)
        ref_attrs = list(result.ref_attrs)

        # special case, the renamed attribute is implicit
        if new_attrs and not ref_attrs:
            if len(new_attrs) != 1:
                raise DataJointError('Renamed foreign key must be mapped to the primary key in "%s"' % line)
            if len(ref.primary_key) == 1:
                # if the primary key has one attribute, allow implicit renaming
                ref_attrs = ref.primary_key
            else:
                # if only one primary key attribute remains, then allow implicit renaming
                ref_attrs = [attr for attr in ref.primary_key if attr not in attributes]
                if len(ref_attrs) != 1:
                    raise DataJointError('Could not resolve which primary key attribute should be referenced in "%s"' % line)

        if len(new_attrs) != len(ref_attrs):
            raise DataJointError('Mismatched attributes in foreign key "%s"' % line)

        if ref_attrs:
            # convert to projected dependency
            ref = ref.proj(**dict(zip(new_attrs, ref_attrs)))

    # declare new foreign key attributes
    for attr in ref.primary_key:
        if attr not in attributes:
            attributes.append(attr)
            if primary_key is not None:
                primary_key.append(attr)
            attr_sql.append(
                ref.heading[attr].sql.replace('NOT NULL ', '', int(is_nullable)))

    # declare the foreign key
    foreign_key_sql.append(
        'FOREIGN KEY (`{fk}`) REFERENCES {ref} (`{pk}`) ON UPDATE CASCADE ON DELETE RESTRICT'.format(
            fk='`,`'.join(ref.primary_key),
            pk='`,`'.join(ref.heading[name].original_name for name in ref.primary_key),
            ref=ref.support[0]))

    # declare unique index
    if is_unique:
        index_sql.append('UNIQUE INDEX ({attrs})'.format(attrs=','.join("`%s`" % attr for attr in ref.primary_key)))


def prepare_declare(definition, context):
    # split definition into lines
    definition = re.split(r'\s*\n\s*', definition.strip())
    # check for optional table comment
    table_comment = definition.pop(0)[1:].strip() if definition[0].startswith('#') else ''
    if table_comment.startswith(':'):
        raise DataJointError('Table comment must not start with a colon ":"')
    in_key = True  # parse primary keys
    primary_key = []
    attributes = []
    attribute_sql = []
    foreign_key_sql = []
    index_sql = []
    external_stores = []

    for line in definition:
        if not line or line.startswith('#'):  # ignore additional comments
            pass
        elif line.startswith('---') or line.startswith('___'):
            in_key = False  # start parsing dependent attributes
        elif is_foreign_key(line):
            compile_foreign_key(line, context, attributes,
                                primary_key if in_key else None,
                                attribute_sql, foreign_key_sql, index_sql)
        elif re.match(r'^(unique\s+)?index[^:]*$', line, re.I):   # index
            compile_index(line, index_sql)
        else:
            name, sql, store = compile_attribute(line, in_key, foreign_key_sql, context)
            if store:
                external_stores.append(store)
            if in_key and name not in primary_key:
                primary_key.append(name)
            if name not in attributes:
                attributes.append(name)
                attribute_sql.append(sql)

    return table_comment, primary_key, attribute_sql, foreign_key_sql, index_sql, external_stores


def declare(full_table_name, definition, context):
    """
    Parse declaration and generate the SQL CREATE TABLE code
    :param full_table_name: full name of the table
    :param definition: DataJoint table definition
    :param context: dictionary of objects that might be referred to in the table
    :return: SQL CREATE TABLE statement, list of external stores used
    """
    table_name = full_table_name.strip('`').split('.')[1]
    if len(table_name) > MAX_TABLE_NAME_LENGTH:
        raise DataJointError(
            'Table name `{name}` exceeds the max length of {max_length}'.format(
                name=table_name,
                max_length=MAX_TABLE_NAME_LENGTH))

    table_comment, primary_key, attribute_sql, foreign_key_sql, index_sql, external_stores = prepare_declare(
        definition, context)

    if not primary_key:
        raise DataJointError('Table must have a primary key')

    return (
        'CREATE TABLE IF NOT EXISTS %s (\n' % full_table_name +
        ',\n'.join(attribute_sql + ['PRIMARY KEY (`' + '`,`'.join(primary_key) + '`)'] + foreign_key_sql + index_sql) +
        '\n) ENGINE=InnoDB, COMMENT "%s"' % table_comment), external_stores


def _make_attribute_alter(new, old, primary_key):
    """
    :param new: new attribute declarations
    :param old: old attribute declarations
    :param primary_key: primary key attributes
    :return: list of SQL ALTER commands
    """
    # parse attribute names
    name_regexp = re.compile(r"^`(?P<name>\w+)`")
    original_regexp = re.compile(r'COMMENT "{\s*(?P<name>\w+)\s*}')
    matched = ((name_regexp.match(d), original_regexp.search(d)) for d in new)
    new_names = dict((d.group('name'), n and n.group('name')) for d, n in matched)
    old_names = [name_regexp.search(d).group('name') for d in old]

    # verify that original names are only used once
    renamed = set()
    for v in new_names.values():
        if v:
            if v in renamed:
                raise DataJointError('Alter attempted to rename attribute {%s} twice.' % v)
            renamed.add(v)

    # verify that all renamed attributes existed in the old definition
    try:
        raise DataJointError(
            "Attribute {} does not exist in the original definition".format(
                next(attr for attr in renamed if attr not in old_names)))
    except StopIteration:
        pass

    # dropping attributes
    to_drop = [n for n in old_names if n not in renamed and n not in new_names]
    sql = ['DROP `%s`' % n for n in to_drop]
    old_names = [name for name in old_names if name not in to_drop]

    # add or change attributes in order
    prev = None
    for new_def, (new_name, old_name) in zip(new, new_names.items()):
        if new_name not in primary_key:
            after = None  # if None, then must include the AFTER clause
            if prev:
                try:
                    idx = old_names.index(old_name or new_name)
                except ValueError:
                    after = prev[0]
                else:
                    if idx >= 1 and old_names[idx - 1] != (prev[1] or prev[0]):
                        after = prev[0]
            if new_def not in old or after:
                sql.append('{command} {new_def} {after}'.format(
                    command=("ADD" if (old_name or new_name) not in old_names else
                             "MODIFY" if not old_name else
                             "CHANGE `%s`" % old_name),
                    new_def=new_def,
                    after="" if after is None else "AFTER `%s`" % after))
        prev = new_name, old_name

    return sql


def _make_foreign_key_alter(new, old):
    """
    :param new: new foreign key declarations
    :param old: old foreign key declarations
    :return: list of SQL ALTER commands
    """

    sql = list()
    for foreign_sql in set(new) - set(old):
        sql.append("ADD {foreign_sql}".format(foreign_sql=foreign_sql))

    return sql


def alter(definition, old_definition, context):
    """
    :param definition: new table definition
    :param old_definition: current table definition
    :param context: the context in which to evaluate foreign key definitions
    :return: string SQL ALTER command, list of new stores used for external storage
    """
    table_comment, primary_key, attribute_sql, foreign_key_sql, index_sql, external_stores = prepare_declare(
        definition, context)
    table_comment_, primary_key_, attribute_sql_, foreign_key_sql_, index_sql_, external_stores_ = prepare_declare(
        old_definition, context)

    # analyze differences between declarations
    sql = list()
    if primary_key != primary_key_:
        raise NotImplementedError('table.alter cannot alter the primary key (yet).')
    if set(foreign_key_sql_) - set(foreign_key_sql):
        raise NotImplementedError('table.alter cannot alter foreign keys (yet).')
    if index_sql != index_sql_:
        raise NotImplementedError('table.alter cannot alter indexes (yet)')
    if attribute_sql != attribute_sql_:
        sql.extend(_make_attribute_alter(attribute_sql, attribute_sql_, primary_key))
    if table_comment != table_comment_:
        sql.append('COMMENT="%s"' % table_comment)
    if set(foreign_key_sql) - set(foreign_key_sql_):
        sql.extend(_make_foreign_key_alter(foreign_key_sql, foreign_key_sql_))
    return sql, [e for e in external_stores if e not in external_stores_]


def compile_index(line, index_sql):
    match = index_parser.parseString(line)
    index_sql.append('{unique} index ({attrs})'.format(
        unique=match.unique,
        attrs=','.join('`%s`' % a for a in match.attr_list)))


def substitute_special_type(match, category, foreign_key_sql, context):
    """
    :param match: dict containing with keys "type" and "comment" -- will be modified in place
    :param category: attribute type category from TYPE_PATTERN
    :param foreign_key_sql: list of foreign key declarations to add to
    :param context: context for looking up user-defined attribute_type adapters
    """
    if category == 'UUID':
        match['type'] = UUID_DATA_TYPE
    elif category == 'INTERNAL_ATTACH':
        match['type'] = 'LONGBLOB'
    elif category in EXTERNAL_TYPES:
        if category == 'FILEPATH' and not _support_filepath_types():
            raise DataJointError("""
            The filepath data type is disabled until complete validation.
            To turn it on as experimental feature, set the environment variable
            {env} = TRUE or upgrade datajoint.
            """.format(env=FILEPATH_FEATURE_SWITCH))
        match['store'] = match['type'].split('@', 1)[1]
        match['type'] = UUID_DATA_TYPE
        foreign_key_sql.append(
            "FOREIGN KEY (`{name}`) REFERENCES `{{database}}`.`{external_table_root}_{store}` (`hash`) "
            "ON UPDATE RESTRICT ON DELETE RESTRICT".format(external_table_root=EXTERNAL_TABLE_ROOT, **match))
    elif category == 'ADAPTED':
        adapter = get_adapter(context, match['type'])
        match['type'] = adapter.attribute_type
        # add attribute type used to comment
        match['comment'] = '{comment};{type};'.format(**match)
        category = match_type(match['type'])
        if category in SPECIAL_TYPES:
            # recursive redefinition from user-defined datatypes.
            substitute_special_type(match, category, foreign_key_sql, context)
    else:
        assert False, 'Unknown special type'


def compile_attribute(line, in_key, foreign_key_sql, context):
    """
    Convert attribute definition from DataJoint format to SQL
    :param line: attribution line
    :param in_key: set to True if attribute is in primary key set
    :param foreign_key_sql: the list of foreign key declarations to add to
    :param context: context in which to look up user-defined attribute type adapterss
    :returns: (name, sql, is_external) -- attribute name and sql code for its declaration
    """
    try:
        match = attribute_parser.parseString(line + '#', parseAll=True)
    except pp.ParseException as err:
        raise DataJointError('Declaration error in position {pos} in line:\n  {line}\n{msg}'.format(
            line=err.args[0], pos=err.args[1], msg=err.args[2]))
    match['comment'] = match['comment'].rstrip('#')
    if 'default' not in match:
        match['default'] = ''
    match = {k: v.strip() for k, v in match.items()}
    match['nullable'] = match['default'].lower() == 'null'

    if match['nullable']:
        if in_key:
            raise DataJointError('Primary key attributes cannot be nullable in line "%s"' % line)
        match['default'] = 'DEFAULT NULL'  # nullable attributes default to null
    else:
        if match['default']:
            quote = (match['default'].split('(')[0].upper() not in CONSTANT_LITERALS
                     and match['default'][0] not in '"\'')
            match['default'] = 'NOT NULL DEFAULT ' + ('"%s"' if quote else "%s") % match['default']
        else:
            match['default'] = 'NOT NULL'

    match['comment'] = match['comment'].replace('"', '\\"')   # escape double quotes in comment

    if match['comment'].startswith(':'):
        raise DataJointError('An attribute comment must not start with a colon in comment "{comment}"'.format(**match))

    category = match_type(match['type'])
    if category in SPECIAL_TYPES:
        match['comment'] = ':{type}:{comment}'.format(**match)  # insert custom type into comment
        substitute_special_type(match, category, foreign_key_sql, context)

    if category in SERIALIZED_TYPES and match['default'] not in {'DEFAULT NULL', 'NOT NULL'}:
        raise DataJointError(
            'The default value for a blob or attachment attributes can only be NULL in:\n{line}'.format(line=line))

    sql = ('`{name}` {type} {default}' + (' COMMENT "{comment}"' if match['comment'] else '')).format(**match)
    return match['name'], sql, match.get('store')



"""Defines settings table class for automaker
"""

import importlib
import inspect
import collections
import warnings
import sys
import pickle
import os
from pathlib import Path
import json
import numpy as np
import pandas

try:
    import cloudpickle
    pickle.dumps = cloudpickle.dumps
except ImportError:
    pass

from .table import FreeTable
from .user_tables import UserTable, _base_regexp
from .expression import QueryExpression
from .utils import ClassProperty, from_camel_case, safe_write
from .errors import DataJointError
from .settings import config, prefix_to_role


Sequence = (collections.MutableSequence, tuple, set)
# which keys to keep in conda list return
CONDA_KEEP = ['name', 'version']
# major python version
PYTHON_VERSION = (sys.version_info[0], sys.version_info[1])


def get_settings_table(child_cls, name=None, assigned_schema=None):
    """
    dynamically create a settings table.
    """
    if name is None:
        name = child_cls.name + 'Settings'
        _save_name = name
        settings_prefix = child_cls.table_name.strip('_#')
    else:
        settings_prefix = from_camel_case(name)
        _save_name = child_cls.name + name

    class Settings(Settingstable):

        @property
        def definition(self):
            return super().definition.format(
                table_name=settings_prefix
            )

        @ClassProperty
        def name(cls):
            return name

        @ClassProperty
        def _save_name(cls):
            return _save_name

        @ClassProperty
        def child_table(cls):
            return child_cls

        @ClassProperty
        def _assigned_schema(cls):
            return assigned_schema

    return Settings


class Settingstable(UserTable):
    """settings table class
    """

    _prefix = r'##'
    tier_regexp = r'(?P<settingstable>' + _prefix + _base_regexp + ')'

    @property
    def definition(self):
        return """
        {table_name}_populate_settings : varchar(63)
        ---
        description = null : varchar(4000) # any string to describe setting
        func : longblob # two-tuple of strings (module, function) or callable
        global_settings : longblob # dictionary
        entry_settings : longblob # dictionary
        fetch_method = 'fetch1' : enum('fetch', 'fetch1', 'farfetch', 'farfetch1')
        fetch_tables = null : longblob # dictionary of dict(table_name: projection)
        assign_output = null : longblob # overrides make_compatible to simply assign the output to a column
        restrictions = null : longblob # dictionary or list of restrictions
        parse_unique = null : longblob # list of unique entries for fetch
        created = CURRENT_TIMESTAMP : timestamp
        """

    @ClassProperty
    def child_table(cls):
        raise NotImplementedError('child table attribute for settings table.')

    @ClassProperty
    def _save_name(cls):
        raise NotImplementedError('save name for settings table.')

    @ClassProperty
    def _assigned_schema(cls):
        return None

    @property
    def settings_name(self):
        return self.heading.primary_key[0]

    @staticmethod
    def _check_settings(settings, params, args, kwargs):
        """check if global/entry settings key in function
        """

        if settings is None:
            return {}

        assert isinstance(settings, collections.Mapping), \
            'global_settings must be dictionary'

        for param, value in settings.items():
            if param not in params:
                if param == args:
                    assert isinstance(value, Sequence), (
                        'variable positional (*args) must be sequence, '
                        'but is {}'.format(type(value)))
                elif param == kwargs:
                    assert isinstance(value, collections.Mapping), (
                        'keyword positional (**kwargs) must be mapping, '
                        'but is {}'.format(type(value))
                    )
                elif kwargs is not None:
                    warnings.warn(
                        'unknown keyword argument is being used, '
                        'but variable keyword (**kwargs) exists.'
                    )
                else:
                    raise DataJointError(
                        'global argument {} not in function'.format(param)
                    )

        return settings

    @staticmethod
    def _used_params(global_settings, entry_settings):
        return list(global_settings) + list(entry_settings)

    def _get_required_proj(self, entry_settings):
        """
        """

        required_proj = []

        for value in entry_settings.values():
            if isinstance(value, str):
                required_proj.append(value)
            elif isinstance(value, Sequence):
                required_proj.extend(value)
            elif isinstance(value, dict):
                required_proj.extend(self._get_required_proj(value))

        return required_proj

    def _check_fetch_tables(self, fetch_tables, required_proj):
        """check if fetch tables are correct and return formatted fetch tables
        """

        if fetch_tables is None:
            # default will be to use just the primary parent tables,
            # excluding settings table
            parent_tables = self.child_table().primary_parents(required_proj)

            if parent_tables is None and not required_proj:
                pass
            else:
                assert parent_tables is not None, \
                    'no parent tables, but required projections.'

                required_left = (
                    set(required_proj) - set(parent_tables.heading.names)
                )
                assert not required_left, \
                    'parent table do not contain all required projections.'

            return

        if isinstance(fetch_tables, QueryExpression):
            warnings.warn(
                'inserting query expression for fetch tables not tested.'
            )

            required_left = (
                set(required_proj) - set(fetch_tables.heading.names)
            )

            assert not required_left, \
                'query expression does not contains all required projections.'

            # insert_fetch_tables = fetch_tables
            insert_fetch_tables = pickle.dumps(
                fetch_tables, protocol=pickle.HIGHEST_PROTOCOL
            )

        elif isinstance(fetch_tables, collections.Mapping):
            assert len(fetch_tables) != 0, \
                'fetch tables cannot be empty mapping, use None instead.'
            # load graph
            self.connection.dependencies.load()
            nodes = self.connection.dependencies.nodes
            # load descendents
            descendants = self.child_table().descendants()

            # initialize new dictionary
            insert_fetch_tables = {}
            # initialize joined table to test
            test_joined_table = None

            for table, proj in fetch_tables.items():
                # if not string assume if table instance or class
                if not isinstance(table, str):
                    table = table.full_table_name
                # assume if ` in table then it is in proper notation
                if table.startswith('`'):
                    if table not in nodes:
                        raise DataJointError(
                            'table {} not in database.'.format(table)
                        )
                elif '.' not in table:
                    raise DataJointError(
                        'schema not specified or not separated with a period '
                        'in table name {}.'.format(table)
                    )
                else:
                    # assumed to be in camel case here.
                    table_splits = table.split('.')
                    schema = table_splits.pop(0)

                    # join part tables if necessary
                    table_name = '__'.join([
                        from_camel_case(s) for s in table_splits
                    ])

                    q = '`'
                    combiner = lambda u: (
                        q + schema + q + '.'
                        + q + u + table_name + q)

                    # check manual, imported and computed, autocomputed, autoimported
                    for prefix in prefix_to_role:
                        if combiner(prefix) in nodes:
                            table = combiner(prefix)
                            break
                    else:
                        raise DataJointError((
                            'When processing fetch tables, '
                            'table {table_name} in {schema} was not found.'
                        ).format(table_name=table_name, schema=schema))

                free_table = FreeTable(self.connection, table)
                if free_table.full_table_name in descendants:
                    raise DataJointError((
                        'Cannot have descendant tables in joined table. '
                        'Table {table_name} is a descendant table of '
                        '{child_table_name}.'
                    ).format(
                        table_name=free_table.full_table_name,
                        child_table_name=self.child_table.full_table_name
                    ))

                if isinstance(proj, tuple):
                    assert len(proj) == 2, 'projection must be two-tuple.'

                    assert isinstance(proj[0], Sequence), \
                        'first tuple must be sequence'

                    assert isinstance(proj[1], collections.Mapping), \
                        'second tuple must be mapping'

                    try:
                        free_table.proj(*proj[0], **proj[1])
                    except DataJointError as e:
                        raise DataJointError((
                            'Unable to project table {table}; error: {e}'
                            ).format(table=table, e=e))
                elif isinstance(proj, Sequence):
                    proj = (proj, {})
                elif isinstance(proj, collections.Mapping):
                    proj = ([], proj)
                else:
                    raise DataJointError(
                        'projection must be of type two-tuple, sequence, '
                        'or mapping, but is of type {}'.format(type(proj))
                    )

                try:
                    proj_table = free_table.proj(*proj[0], **proj[1])
                except DataJointError as e:
                    raise DataJointError((
                        'Unable to project table {table}; error: {e}'
                        ).format(table=table, e=e))

                if test_joined_table is None:
                    test_joined_table = proj_table
                else:
                    test_joined_table = test_joined_table * proj_table

                insert_fetch_tables[table] = proj

            required_left = (
                set(required_proj) - set(test_joined_table.heading.names)
            )

            assert not required_left, \
                'joined table does not contain all required projections.'

        return insert_fetch_tables

    def _get_joined_table(self, fetch_tables, required_proj, restrictions):
        """convert fetch_tables to joined table and set fetch_tables_attribute
        """

        if fetch_tables is None:
            # default will be to use just the primary parent tables,
            # excluding settings table
            parent_tables = self.child_table().primary_parents(
                required_proj, restrictions
            )
            return parent_tables

        elif isinstance(fetch_tables, QueryExpression):
            if restrictions is None:
                return fetch_tables
            else:
                return fetch_tables & restrictions

        elif isinstance(fetch_tables, bytes):
            fetch_tables = pickle.loads(fetch_tables)
            if restrictions is None:
                return fetch_tables
            else:
                return fetch_tables & restrictions

        else:
            # load graph
            self.connection.dependencies.load()
            nodes = self.connection.dependencies.nodes

            joined_table = None

            for table, proj in fetch_tables.items():

                if table not in nodes:
                    raise DataJointError(
                        'previously existing table '
                        '{} has been removed'.format(table)
                    )

                proj_table = FreeTable(
                    self.connection, table
                ).proj(*proj[0], **proj[1])

                if joined_table is None:
                    joined_table = proj_table
                else:
                    joined_table = joined_table * proj_table

            if restrictions is None:
                return joined_table
            else:
                return joined_table & restrictions

    @staticmethod
    def _get_func(func):
        """get function
        """

        def func_from_tuple(func):
            """get function from tuple
            """

            module = func[0]
            func = func[1]

            # use importlib to import module
            try:
                module = importlib.import_module(module)
                return getattr(module, func)
            except Exception as e:
                raise DataJointError(
                    'could not load function: {}'.format(e))

        if isinstance(func, (tuple, list)):
            if len(func) == 4:
                # if tuple is of length four it is considered a class
                # with initialization
                cls = func_from_tuple(func)
                args = func[2]
                kwargs = func[3]
                assert isinstance(args, Sequence)
                assert isinstance(args, collections.Mapping)
                assert hasattr(cls, '__init__')
                func = cls(*args, **kwargs)
            elif len(func) == 2:
                # here it is simply considered a function
                if isinstance(func[0], bytes):
                    # in this case it is a python file that was saved
                    if config['tmp_folder'] is None:
                        tmp_folder = os.path.abspath('.')
                    else:
                        tmp_folder = os.path.abspath(config['tmp_folder'])

                    filename, content = func[0].split(b'\0', 1)
                    filename = filename.decode()
                    content = content.decode()

                    temporary_path = Path(tmp_folder) / filename

                    safe_write(temporary_path, content)

                    func = (
                        os.path.splitext(filename)[0],
                        func[1]
                    )

                    # add to sys path if necessary
                    if tmp_folder not in sys.path:
                        try:
                            # put after current directory
                            index = sys.path.index('')
                            sys.path.insert(index+1, tmp_folder)
                        except ValueError:
                            sys.path.append(tmp_folder)
                    func = func_from_tuple(func)

                else:
                    # a normal function was saved
                    func = func_from_tuple(func)
            else:
                raise DataJointError(
                    'tuple must have two or four '
                    'elements, it has {}'.format(len(func))
                )
        elif isinstance(func, bytes):
            # loading pickle function
            func = pickle.loads(func)

        return func

    @staticmethod
    def _check_func(func):

        if (
            not hasattr(func, '__call__')
            or inspect.isclass(func)
            or inspect.ismodule(func)
        ):
            raise DataJointError(
                'function must be two/four-tuple or callable function.'
            )

        module = inspect.getmodule(func)

        if module is None:
            raise DataJointError('module for function is None.')
        elif not hasattr(module, '__file__'):
            raise DataJointError(
                'module {} does not have an associated file'.format(module)
            )

    @staticmethod
    def _get_func_status(func):
        """
        get version and package for function
        get git status for function if it exists
        get conda environment list
        """

        # init func status dictionary
        func_status = {}
        func_status['python_version'] = PYTHON_VERSION

        module = _get_package_status(func_status, func)

        _get_conda_status(func_status)

        _get_git_status(func_status, module, func)

        return func_status

    def _check_func_status(self, func_dict):
        """
        check version and package for function
        check git status between inserted and current version
        check anaconda environment list
        """

        func = func_dict['func']
        func_status = self._get_func_status(func)

        new_sha1 = func_status.get('sha1', None)
        new_branch = func_status.get('branch', None)
        new_modified = func_status.get('modified', None)
        new_version = func_status.get('version', None)
        new_package = func_status.get('package', None)
        new_conda_list = func_status.get('conda_list', None)
        new_python_version = func_status.get('python_version', None)

        old_sha1 = func_dict.get('sha1', None)
        old_branch = func_dict.get('branch', None)
        old_modified = func_dict.get('modified', None)
        old_version = func_dict.get('version', None)
        old_package = func_dict.get('package', None)
        old_conda_list = func_dict.get('conda_list', None)
        old_python_version = func_dict.get('python_version', None)

        # check python version
        if (new_python_version is None) or (old_python_version is None):
            if config['automaker_warnings']:
                warnings.warn((
                    'No python version checking for function '
                    '"{func}" available.'
                ).format(func=func))
        elif new_python_version != old_python_version:
            if config['automaker_warnings']:
                warnings.warn((
                    'Old python version "{old_python_version}" '
                    'does not match current python '
                    'version "{new_python_version}"'
                ).format(
                    new_python_version=new_python_version,
                    old_python_version=old_python_version
                ))

        _check_package_status(
            func,
            new_package, old_package,
            new_version, old_version
        )

        _check_conda_status(
            func,
            new_conda_list, old_conda_list
        )

        _check_git_status(
            new_sha1, old_sha1,
            old_branch, new_branch,
            new_modified, old_modified
        )

    @staticmethod
    def _get_func_params(func):
        """get available function parameters
        """

        signature = inspect.signature(func)

        params = {}
        args = None
        kwargs = None
        pos_or_kw = False
        for arg, param in signature.parameters.items():

            if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
                pos_or_kw = True

            if param.kind is inspect.Parameter.POSITIONAL_ONLY:
                raise DataJointError(
                    'function cannot have position only argument.'
                    'case for: {}'.format(arg)
                )
            if param.kind is inspect.Parameter.VAR_POSITIONAL:
                args = arg
            if param.kind is inspect.Parameter.VAR_KEYWORD:
                kwargs = arg
            elif param.default is inspect._empty:
                params[arg] = None
            else:
                params[arg] = param.default

        if pos_or_kw and args is not None:
            raise DataJointError(
                'function cannot have arguments before '
                'variable positional (*args).'
            )

        return {
            'params': params,
            'args': args,
            'kwargs': kwargs}

    @staticmethod
    def _check_func_params(func_dict, used):
        """check function parameters

        :param func: callable function
        :param params: dictionary of inserted parameters
        :param args: inserted argument that is a variable positional (*args).
        :param kwargs: inserted argument that is the variable keyword (**kwargs).
        :param used: arguments used for function.
        """

        func = func_dict['func']
        params = func_dict.get('params', {})
        args = func_dict.get('args', None)
        kwargs = func_dict.get('kwargs', None)

        signature = inspect.signature(func)

        dictionary = dict(signature.parameters)

        if args is not None and args in used:
            try:
                dictionary.pop(args)
            except KeyError:
                raise DataJointError(
                    'function does not contain the '
                    '"{}" variable argument anymore.'.format(args))
        if kwargs is not None and kwargs in used:
            try:
                dictionary.pop(kwargs)
            except KeyError:
                raise DataJointError(
                    'function does not contain the '
                    '"{}" variable keyword argument anymore.'.format(kwargs))

        for param, default in params.items():
            # check if
            if param not in dictionary and param in used:
                raise DataJointError(
                    'function does not contain the '
                    '"{}" argument anymore'.format(param)
                )

            if (default != dictionary[param].default) and param not in used:
                raise DataJointError(
                    'unused argument '
                    '"{}" has changed its default value'.format(param)
                )

    @staticmethod
    def _get_insert_func(func):
        """get function ready for insertion
        """

        if isinstance(func, (tuple, list)):
            if len(func) == 2:
                # if python file encode it into bytes
                if func[0].endswith('.py'):
                    if not os.path.exists(func[0]):
                        raise DataJointError(
                            'python file {} does not exist.'.format(func[0])
                        )
                    # init path object
                    python_file = Path(func[0])

                    func = ((
                        str.encode(python_file.name)
                        + b'\0'
                        + python_file.read_bytes()
                    ), func[1])
        else:
            # make binary
            func = pickle.dumps(func, protocol=pickle.HIGHEST_PROTOCOL)

        return func

    def _get_func_attr(self, func):
        """convert to function attribute and check function
        """

        attr = {'func': self._get_insert_func(func)}

        func = self._get_func(attr['func'])
        self._check_func(func)
        attr.update(self._get_func_status(func))
        attr.update(self._get_func_params(func))

        return attr

    @staticmethod
    def _check_restrictions(restrictions):
        """check if restrictions are list or dict
        """

        if restrictions is None:
            pass
        elif not isinstance(restrictions, (list, dict, np.recarray, str)):
            raise DataJointError(
                'constant restriction for insertion must be '
                'list, dict, or recarray (or string).'
            )

    @staticmethod
    def _check_parse_unique(parse_unique, required_proj):
        """check parse unique
        """

        if parse_unique is None:
            pass
        elif not isinstance(parse_unique, Sequence):
            raise DataJointError(
                'the parse unique attribute must be a sequence.'
            )
        else:
            left_parse = set(parse_unique) - set(required_proj)

            assert not left_parse, (
                'parse may only contain elements that are attributes '
                'to be fetched. Also contains {left_parse}, and not '
                'just {required_proj}'
            ).format(left_parse=left_parse, required_proj=required_proj)

    def _check_assign_output(self, row):

        assign_output = row.get('assign_output', None)

        if (
            assign_output is not None
            and assign_output
            not in self.child_table.heading.secondary_attributes
        ):
            raise DataJointError((
                '"{}" not in secondary columns of table "{}".'
            ).format(assign_output, self.child_table.full_table_name))

    def insert(self, *args, **kwargs):
        raise NotImplementedError(
            'For a Settingstable class only'
            ' insert1 is implemented.')

    def insert1(self, row, **kwargs):

        # aliases for primary key
        primary_key = self.settings_name
        truth = primary_key in row
        if not truth and 'settings_name' in row:
            row[primary_key] = row.pop('settings_name')
        elif not truth and 'settings' in row:
            row[primary_key] = row.pop('settings')
        elif not truth and 'populate_settings' in row:
            row[primary_key] = row.pop('populate_settings')
        elif not truth:
            raise DataJointError(
                'Need to provide a name for settings; you can use the '
                'following key aliases in your dictionary for insertion: '
                '"settings_name", "settings", and "populate_settings".'
            )

        # remove inferred/default attributes
        row.pop('args', None)
        row.pop('kwargs', None)
        row.pop('created', None)

        row['func'] = self._get_func_attr(row['func'])

        row['global_settings'] = self._check_settings(
            row.get('global_settings', None),
            row['func']['params'],
            row['func']['args'],
            row['func']['kwargs']
        )

        row['entry_settings'] = self._check_settings(
            row.get('entry_settings', None),
            row['func']['params'],
            row['func']['args'],
            row['func']['kwargs']
        )

        # required to be contained in the final joined table
        required_proj = self._get_required_proj(row['entry_settings'])

        row['fetch_tables'] = self._check_fetch_tables(
            row.get('fetch_tables', None), required_proj)

        self._check_restrictions(row.get('restrictions', None))
        self._check_parse_unique(row.get('parse_unique', None), required_proj)
        self._check_assign_output(row)

        # not implemented farfetch
        if 'farfetch' in row.get('fetch_method', 'fetch1'):
            raise NotImplementedError('farfetch method.')

        return super().insert((row,), **kwargs)

    def fetch1(
        self, *args,
        check_function=True, get_joined_table=True,
        **kwargs
    ):

        row = super().fetch1(*args, **kwargs)

        return self._postfetch_processing(
            row, check_function, get_joined_table
        )

    def fetch(self, *args, **kwargs):
        warnings.warn('fetch in Settingstable class does not process each '
                      'tuple, use fetch1 instead.')
        return super().fetch(*args, **kwargs)

    def _postfetch_processing(
        self, row, check_function=True, get_joined_table=True
    ):
        # convert and check function
        row['func']['func'] = self._get_func(row['func']['func'])
        if check_function:
            self._check_func(row['func']['func'])
            self._check_func_status(row['func'])
            self._check_func_params(
                row['func'],
                used=self._used_params(
                    row['global_settings'], row['entry_settings']
                ),
            )
        row['args'] = row['func'].get('args', None)
        row['kwargs'] = row['func'].get('kwargs', None)
        row['func'] = row['func']['func']

        if get_joined_table:
            required_proj = self._get_required_proj(row['entry_settings'])
            # get joined tables / primary parent tables
            row['fetch_tables'] = self._get_joined_table(
                row['fetch_tables'],
                required_proj,
                row['restrictions']
            )

        return row


# -- misc helper functions -- #

def _get_package_status(func_status, func):
    """
    get package status (used by Settingstable._get_func_status)
    returns module
    """

    # attribute file check for module happend previously
    module = inspect.getmodule(func)

    # check if module is package
    if hasattr(module, '__package__'):
        # get package module
        module_name = getattr(module, '__package__').split('.')[0]
        func_status['package'] = module_name

        # __main__ may be skipped
        try:
            # load package module
            module = importlib.import_module(module_name)
            # possibly redundant
            module = inspect.getmodule(module)
        except ValueError:
            pass

    # check if module has a version
    if hasattr(module, '__version__'):
        func_status['version'] = getattr(module, '__version__')

    return module


def _check_package_status(
    func,
    new_package, old_package,
    new_version, old_version
):
    if not config['automaker_warnings']:
        return
    # check package and package version
    if (new_package is None) or (old_package is None):
        warnings.warn((
            'No package checking for function "{func}" available.'
        ).format(func=func))
    elif new_package != old_package:
        warnings.warn((
            'Old package "{old_package}" does not '
            'match with new package "{new_package}".'
        ).format(old_package=old_package, new_package=new_package))
    elif (new_version is None) or (old_version is None):
        warnings.warn((
            'No version checking for package "{package}" available.'
        ).format(package=old_package))
    elif new_version != old_version:
        warnings.warn((
            'Old version "{old_version}" for package "{package}" '
            'does not match new version "{new_version}"'
        ).format(
            package=old_package,
            old_version=old_version,
            new_version=new_version
        ))


def _get_conda_status(func_status):
    """
    get conda status (used by Settingstable._get_func_status)
    """

    # try importing conda
    try:
        import conda.cli.python_api as conda_api
    except ImportError:
        if config['automaker_warnings']:
            warnings.warn(
                'Did not perform getting conda status: '
                'Conda Python API not installed.'
            )
        return

    try:
        stdout, stderr, return_code = conda_api.run_command(
            conda_api.Commands.LIST, '--json', use_exception_handler=True
        )
    except Exception:
        if config['automaker_warnings']:
            warnings.warn(
                'Could not perform "conda list --json". '
                'No conda status test performed'
            )
        return

    if return_code:
        if config['automaker_warnings']:
            warnings.warn((
                'Could not get conda environment list; Error message:\n{stderr}'
                '\n{error}.'
            ).format(stderr=stderr, error=stdout['error']))
        return

    conda_list = pandas.DataFrame(json.loads(stdout))[CONDA_KEEP]

    func_status['conda_list'] = conda_list.to_dict('list')


def _check_conda_status(
    func,
    new_conda_list, old_conda_list
):
    if not config['automaker_warnings']:
        return
    # check conda list
    if (new_conda_list is None) or (old_conda_list is None):
        warnings.warn((
            'No conda checking for function "{func}" available.'
        ).format(func=func))
    elif new_conda_list != old_conda_list:
        new_conda_list = pandas.DataFrame(new_conda_list)
        old_conda_list = pandas.DataFrame(old_conda_list)
        # only use keys from old conda list
        # i.e. it doesn't care if new packages were installed
        merged_conda_list = pandas.merge(
            new_conda_list, old_conda_list,
            on='name',
            how='right',
            suffixes=('_new', '_old')
        )
        # packages missing in new version
        missing_packages = merged_conda_list['version_new'].isnull()
        if missing_packages.any():
            warnings.warn((
                'These packages, which were present when inserting '
                'this function, are missing in your current anaconda '
                'environment:\n{missing}'
            ).format(
                missing=merged_conda_list.loc[
                    missing_packages, 'name'
                ].tolist()
            ))
        # packages with different versions
        different_version = (
            merged_conda_list['version_old']
            != merged_conda_list['version_new']
        )
        if different_version.any():
            merged_conda_list = merged_conda_list[different_version]
            warnings.warn((
                'These packages have different versions from the time '
                'this function was inserted into the Settingstable:'
                '\n{different}'
            ).format(
                different=merged_conda_list
            ))


def _get_git_status(func_status, module, func):
    """
    get git status (used by Settingstable._get_func_status)
    """

    # try importing git module for git checking
    try:
        import git
    except ImportError:
        if config['automaker_warnings']:
            warnings.warn(
                'Did not perform getting git status: '
                'Git Python API not installed.')
        return

    # get directory name of module
    dir_name = os.path.dirname(inspect.getabsfile(module))
    module_path = dir_name

    while True:
        # set git path
        git_path = os.path.join(dir_name, '.git')
        # set git repo if path exists
        if os.path.exists(git_path):
            repo = git.Repo(git_path)
            sha1, branch = repo.head.commit.name_rev.split()
            # check if files were modified
            modified = (repo.git.status().find('modified') > 0)
            if modified and config['automaker_warnings']:
                warnings.warn(
                    'You have uncommited changes. '
                    'Consider committing before running populate.'
                )

            func_status.update({
                'sha1': sha1,
                'branch': branch,
                'modified': modified,
            })

            break

        parts = os.path.split(dir_name)

        if dir_name in parts:
            # only throw a warning if not package and versioning exists
            if (
                'package' not in func_status
                and 'version' not in func_status
                and config['automaker_warnings']
            ):
                warnings.warn((
                    'No git directory was found for module in {path} for '
                    'function {func}. '
                    'Implementation of git version control recommended.'
                ).format(path=module_path, func=func))

            break

        dir_name = parts[0]


def _check_git_status(
    new_sha1, old_sha1,
    old_branch, new_branch,
    new_modified, old_modified
):
    if not config['automaker_warnings']:
        return
    # check git commit
    if (new_sha1 is None) or (old_sha1 is None):
        return
    if old_branch != new_branch:
        warnings.warn((
            'Working in new branch {new_branch}. '
            'Old branch was {old_branch}'
        ).format(new_branch=new_branch, old_branch=old_branch))
    if new_sha1 != old_sha1:
        warnings.warn(
            'Git commits have occured since insertion.'
        )
    elif new_modified and not old_modified:
        warnings.warn(
            'Files have been modified since insertion.'
        )



"""Defines auto make class and settings table
Contains AutoComputed and AutoImported tables
"""

import collections

from .user_tables import UserTable, _base_regexp
from .automake import AutoMake

Sequence = (collections.MutableSequence, tuple, set)


class AutoImported(UserTable, AutoMake):
    """
    Inherit from this class if the table's values are imported from external
    data sources.
    """

    _prefix = '#_'
    tier_regexp = r'(?P<autoimported>' + _prefix + _base_regexp + ')'

    def __init__(self, *args, **kwargs):

        self.set_true_definition()
        super().__init__(*args, **kwargs)


class AutoComputed(UserTable, AutoMake):
    """
    Inherit from this class if the table's values are computed from other
    relations in the schema.
    """

    _prefix = '_#'
    tier_regexp = r'(?P<autocomputed>' + _prefix + _base_regexp + ')'

    def __init__(self, *args, **kwargs):

        self.set_true_definition()
        super().__init__(*args, **kwargs)



import hashlib
import uuid
import io
from pathlib import Path


def key_hash(mapping):
    """
    32-byte hash of the mapping's key values sorted by the key name.
    This is often used to convert a long primary key value into a shorter hash.
    For example, the JobTable in datajoint.jobs uses this function to hash the primary key of autopopulated tables.
    """
    hashed = hashlib.md5()
    for k, v in sorted(mapping.items()):
        hashed.update(str(v).encode())
    return hashed.hexdigest()


def uuid_from_stream(stream, *, init_string=""):
    """
    :return: 16-byte digest of stream data
    :stream: stream object or open file handle
    :init_string: string to initialize the checksum
    """
    hashed = hashlib.md5(init_string.encode())
    chunk = True
    chunk_size = 1 << 14
    while chunk:
        chunk = stream.read(chunk_size)
        hashed.update(chunk)
    return uuid.UUID(bytes=hashed.digest())


def uuid_from_buffer(buffer=b"", *, init_string=""):
    return uuid_from_stream(io.BytesIO(buffer), init_string=init_string)


def uuid_from_file(filepath, *, init_string=""):
    with Path(filepath).open("rb") as f:
        uuid_obj = uuid_from_stream(f, init_string=init_string)
    return uuid_obj



import datajoint as dj
from pathlib import Path
import re
from .utils import user_choice


def migrate_dj011_external_blob_storage_to_dj012(migration_schema, store):
    """
    Utility function to migrate external blob data from 0.11 to 0.12.
    :param migration_schema: string of target schema to be migrated
    :param store: string of target dj.config['store'] to be migrated
    """
    if not isinstance(migration_schema, str):
        raise ValueError(
            'Expected type {} for migration_schema, not {}.'.format(
                str, type(migration_schema)))

    do_migration = False
    do_migration = user_choice(
            """
Warning: Ensure the following are completed before proceeding.
- Appropriate backups have been taken,
- Any existing DJ 0.11.X connections are suspended, and
- External config has been updated to new dj.config['stores'] structure.
Proceed?
            """, default='no') == 'yes'
    if do_migration:
        _migrate_dj011_blob(dj.Schema(migration_schema), store)
        print('Migration completed for schema: {}, store: {}.'.format(
                migration_schema, store))
        return
    print('No migration performed.')


def _migrate_dj011_blob(schema, default_store):
    query = schema.connection.query

    LEGACY_HASH_SIZE = 43

    legacy_external = dj.FreeTable(
        schema.connection,
        '`{db}`.`~external`'.format(db=schema.database))

    # get referencing tables
    refs = [{k.lower(): v for k, v in elem.items()} for elem in query("""
    SELECT concat('`', table_schema, '`.`', table_name, '`')
            as referencing_table, column_name, constraint_name
    FROM information_schema.key_column_usage
    WHERE referenced_table_name="{tab}" and referenced_table_schema="{db}"
    """.format(
        tab=legacy_external.table_name,
        db=legacy_external.database), as_dict=True).fetchall()]

    for ref in refs:
        # get comment
        column = query(
            'SHOW FULL COLUMNS FROM {referencing_table}'
            'WHERE Field="{column_name}"'.format(
                **ref), as_dict=True).fetchone()

        store, comment = re.match(
            r':external(-(?P<store>.+))?:(?P<comment>.*)',
            column['Comment']).group('store', 'comment')

        # get all the hashes from the reference
        hashes = {x[0] for x in query(
            'SELECT `{column_name}` FROM {referencing_table}'.format(
                **ref))}

        # sanity check make sure that store suffixes match
        if store is None:
            assert all(len(_) == LEGACY_HASH_SIZE for _ in hashes)
        else:
            assert all(_[LEGACY_HASH_SIZE:] == store for _ in hashes)

        # create new-style external table
        ext = schema.external[store or default_store]

        # add the new-style reference field
        temp_suffix = 'tempsub'

        try:
            query("""ALTER TABLE {referencing_table}
                ADD COLUMN `{column_name}_{temp_suffix}` {type} DEFAULT NULL
            COMMENT ":blob@{store}:{comment}"
            """.format(
                type=dj.declare.UUID_DATA_TYPE,
                temp_suffix=temp_suffix,
                store=(store or default_store), comment=comment, **ref))
        except:
            print('Column already added')
            pass

        for _hash, size in zip(*legacy_external.fetch('hash', 'size')):
            if _hash in hashes:
                relative_path = str(Path(schema.database, _hash).as_posix())
                uuid = dj.hash.uuid_from_buffer(init_string=relative_path)
                external_path = ext._make_external_filepath(relative_path)
                if ext.spec['protocol'] == 's3':
                    contents_hash = dj.hash.uuid_from_buffer(ext._download_buffer(external_path))
                else:
                    contents_hash = dj.hash.uuid_from_file(external_path)
                ext.insert1(dict(
                    filepath=relative_path,
                    size=size,
                    contents_hash=contents_hash,
                    hash=uuid
                ), skip_duplicates=True)

                query(
                    'UPDATE {referencing_table} '
                    'SET `{column_name}_{temp_suffix}`=%s '
                    'WHERE `{column_name}` = "{_hash}"'
                    .format(
                        _hash=_hash,
                        temp_suffix=temp_suffix, **ref), uuid.bytes)

        # check that all have been copied
        check = query(
            'SELECT * FROM {referencing_table} '
            'WHERE `{column_name}` IS NOT NULL'
            '  AND `{column_name}_{temp_suffix}` IS NULL'
            .format(temp_suffix=temp_suffix, **ref)).fetchall()

        assert len(check) == 0, 'Some hashes havent been migrated'

        # drop old foreign key, rename, and create new foreign key
        query("""
            ALTER TABLE {referencing_table}
            DROP FOREIGN KEY `{constraint_name}`,
            DROP COLUMN `{column_name}`,
            CHANGE COLUMN `{column_name}_{temp_suffix}` `{column_name}`
                {type} DEFAULT NULL
                COMMENT ":blob@{store}:{comment}",
            ADD FOREIGN KEY (`{column_name}`) REFERENCES {ext_table_name}
                (`hash`)
            """.format(
                temp_suffix=temp_suffix,
                ext_table_name=ext.full_table_name,
                type=dj.declare.UUID_DATA_TYPE,
                store=(store or default_store), comment=comment, **ref))

    # Drop the old external table but make sure it's no longer referenced
    # get referencing tables
    refs = [{k.lower(): v for k, v in elem.items()} for elem in query("""
    SELECT concat('`', table_schema, '`.`', table_name, '`') as
        referencing_table, column_name, constraint_name
    FROM information_schema.key_column_usage
    WHERE referenced_table_name="{tab}" and referenced_table_schema="{db}"
    """.format(
        tab=legacy_external.table_name,
        db=legacy_external.database), as_dict=True).fetchall()]

    assert not refs, 'Some references still exist'

    # drop old external table
    legacy_external.drop_quick()



"""Part table class
"""

from .utils import from_camel_case, ClassProperty
from .errors import DataJointError
from .user_tables import UserTable, Manual, Lookup, Imported, Computed, \
    _base_regexp
from .auto_tables import AutoComputed, AutoImported


class Part(UserTable):
    """
    Inherit from this class if the table's values are details of an entry in another relation
    and if this table is populated by this relation. For example, the entries inheriting from
    dj.Part could be single entries of a matrix, while the parent table refers to the entire matrix.
    Part relations are implemented as classes inside classes.
    """

    _connection = None
    _master = None

    tier_regexp = r'(?P<master>' + '|'.join(
        [c.tier_regexp for c in (Manual, Lookup, Imported, Computed, AutoComputed, AutoImported)]
    ) + r'){1,1}' + '__' + r'(?P<part>' + _base_regexp + ')'

    @ClassProperty
    def connection(cls):
        return cls._connection

    @ClassProperty
    def full_table_name(cls):
        return None if cls.database is None or cls.table_name is None else r"`{0:s}`.`{1:s}`".format(
            cls.database, cls.table_name)

    @ClassProperty
    def master(cls):
        return cls._master

    @ClassProperty
    def table_name(cls):
        return None if cls.master is None else cls.master.table_name + '__' + from_camel_case(cls.name)

    def delete(self, force=False):
        """
        unless force is True, prohibits direct deletes from parts.
        """
        if force:
            super().delete()
        else:
            raise DataJointError('Cannot delete from a Part directly.  Delete from master instead')

    def drop(self, force=False):
        """
        unless force is True, prohibits direct deletes from parts.
        """
        if force:
            super().drop()
        else:
            raise DataJointError('Cannot drop a Part directly.  Delete from master instead')



#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path
import sys

min_py_version = (3, 6)

if sys.version_info <  min_py_version:
    sys.exit('DataJoint is only supported for Python {}.{} or higher'.format(*min_py_version))

here = path.abspath(path.dirname(__file__))

long_description = "A relational data framework for scientific data pipelines with MySQL backend."

# read in version number into __version__
with open(path.join(here, 'datajoint', 'version.py')) as f:
    exec(f.read())

with open(path.join(here, 'requirements.txt')) as f:
    requirements = f.read().split()

setup(
    name='datajoint-gucky',
    version=__version__,
    description="A relational data pipeline framework.",
    long_description=long_description,
    author='Matthias Christenson (forked), Dimitri Yatsenko',
    author_email='gucky@gucky.eu (forked), info@datajoint.io',
    license="GNU LGPL",
    url='https://datajoint.io',
    keywords='database organization',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=requirements,
    python_requires='~={}.{}'.format(*min_py_version),
    setup_requires=['setuptools_certificate'], # maybe remove due to conflicts?
    pubkey_path='./datajoint.pub'
)



import os
from .hash import key_hash
import platform
from .table import Table
from .settings import config
from .errors import DuplicateError
from .heading import Heading

ERROR_MESSAGE_LENGTH = 2047
TRUNCATION_APPENDIX = '...truncated'


class JobTable(Table):
    """
    A base relation with no definition. Allows reserving jobs
    """
    def __init__(self, conn, database):
        self.database = database
        self._connection = conn
        self._heading = Heading(table_info=dict(
            conn=conn,
            database=database,
            table_name=self.table_name,
            context=None
        ))
        self._support = [self.full_table_name]

        self._definition = """    # job reservation table for `{database}`
        table_name  :varchar(255)  # className of the table
        key_hash  :char(32)  # key hash
        ---
        status  :enum('reserved','error','ignore')  # if tuple is missing, the job is available
        key=null  :blob  # structure containing the key
        error_message=""  :varchar({error_message_length})  # error message returned if failed
        error_stack=null  :blob  # error stack if failed
        user="" :varchar(255) # database user
        host=""  :varchar(255)  # system hostname
        pid=0  :int unsigned  # system process id
        connection_id = 0  : bigint unsigned          # connection_id()
        timestamp=CURRENT_TIMESTAMP  :timestamp   # automatic timestamp
        """.format(database=database, error_message_length=ERROR_MESSAGE_LENGTH)
        if not self.is_declared:
            self.declare()
        self._user = self.connection.get_user()

    @property
    def definition(self):
        return self._definition

    @property
    def table_name(self):
        return '~jobs'

    def delete(self):
        """bypass interactive prompts and dependencies"""
        self.delete_quick()

    def drop(self):
        """bypass interactive prompts and dependencies"""
        self.drop_quick()

    def reserve(self, table_name, key):
        """
        Reserve a job for computation.  When a job is reserved, the job table contains an entry for the
        job key, identified by its hash. When jobs are completed, the entry is removed.
        :param table_name: `database`.`table_name`
        :param key: the dict of the job's primary key
        :return: True if reserved job successfully. False = the jobs is already taken
        """
        job = dict(
            table_name=table_name,
            key_hash=key_hash(key),
            status='reserved',
            host=platform.node(),
            pid=os.getpid(),
            connection_id=self.connection.connection_id,
            key=key,
            user=self._user)
        try:
            with config(enable_python_native_blobs=True):
                self.insert1(job, ignore_extra_fields=True)
        except DuplicateError:
            return False
        return True

    def complete(self, table_name, key):
        """
        Log a completed job.  When a job is completed, its reservation entry is deleted.
        :param table_name: `database`.`table_name`
        :param key: the dict of the job's primary key
        """
        job_key = dict(table_name=table_name, key_hash=key_hash(key))
        (self & job_key).delete_quick()

    def error(self, table_name, key, error_message, error_stack=None):
        """
        Log an error message.  The job reservation is replaced with an error entry.
        if an error occurs, leave an entry describing the problem
        :param table_name: `database`.`table_name`
        :param key: the dict of the job's primary key
        :param error_message: string error message
        :param error_stack: stack trace
        """
        if len(error_message) > ERROR_MESSAGE_LENGTH:
            error_message = error_message[:ERROR_MESSAGE_LENGTH-len(TRUNCATION_APPENDIX)] + TRUNCATION_APPENDIX
        with config(enable_python_native_blobs=True):
            self.insert1(
                dict(
                    table_name=table_name,
                    key_hash=key_hash(key),
                    status="error",
                    host=platform.node(),
                    pid=os.getpid(),
                    connection_id=self.connection.connection_id,
                    user=self._user,
                    key=key,
                    error_message=error_message,
                    error_stack=error_stack),
                replace=True, ignore_extra_fields=True)



"""
AWS S3 operations
"""
from io import BytesIO
import minio   # https://docs.minio.io/docs/python-client-api-reference
import warnings
import uuid
import logging
from pathlib import Path
from . import errors

logger = logging.getLogger(__name__)


class Folder:
    """
    A Folder instance manipulates a flat folder of objects within an S3-compatible object store
    """
    def __init__(self, endpoint, bucket, access_key, secret_key, *, secure=False, **_):
        self.client = minio.Minio(endpoint, access_key=access_key, secret_key=secret_key,
                                  secure=secure)
        self.bucket = bucket
        if not self.client.bucket_exists(bucket):
            raise errors.BucketInaccessible('Inaccessible s3 bucket %s' % bucket)

    def put(self, name, buffer):
        logger.debug('put: {}:{}'.format(self.bucket, name))
        return self.client.put_object(
            self.bucket, str(name), BytesIO(buffer), length=len(buffer))

    def fput(self, local_file, name, metadata=None):
        logger.debug('fput: {} -> {}:{}'.format(self.bucket, local_file, name))
        return self.client.fput_object(
            self.bucket, str(name), str(local_file), metadata=metadata)

    def get(self, name):
        logger.debug('get: {}:{}'.format(self.bucket, name))
        try:
            return self.client.get_object(self.bucket, str(name)).data
        except minio.error.S3Error as e:
            if e.code == 'NoSuchKey':
                raise errors.MissingExternalFile('Missing s3 key %s' % name)
            else:
                raise e

    def fget(self, name, local_filepath):
        """get file from object name to local filepath"""
        logger.debug('fget: {}:{}'.format(self.bucket, name))
        name = str(name)
        stat = self.client.stat_object(self.bucket, name)
        meta = {k.lower().lstrip('x-amz-meta'): v for k, v in stat.metadata.items()}
        data = self.client.get_object(self.bucket, name)
        local_filepath = Path(local_filepath)
        local_filepath.parent.mkdir(parents=True, exist_ok=True)
        with local_filepath.open('wb') as f:
            for d in data.stream(1 << 16):
                f.write(d)
        if 'contents_hash' in meta:
            return uuid.UUID(meta['contents_hash'])

    def exists(self, name):
        logger.debug('exists: {}:{}'.format(self.bucket, name))
        try:
            self.client.stat_object(self.bucket, str(name))
        except minio.error.S3Error as e:
            if e.code == 'NoSuchKey':
                return False
            else:
                raise e
        return True

    def get_size(self, name):
        logger.debug('get_size: {}:{}'.format(self.bucket, name))
        try:
            return self.client.stat_object(self.bucket, str(name)).size
        except minio.error.S3Error as e:
            if e.code == 'NoSuchKey':
                raise errors.MissingExternalFile
            else:
                raise e

    def remove_object(self, name):
        logger.debug('remove_object: {}:{}'.format(self.bucket, name))
        try:
            self.client.remove_object(self.bucket, str(name))
        except minio.ResponseError:
            return errors.DataJointError('Failed to delete %s from s3 storage' % name)



"""
Hosts the table tiers, user relations should be derived from.
"""

import datajoint as dj
from .table import Table
from .autopopulate import AutoPopulate
from .utils import from_camel_case, ClassProperty
from .errors import DataJointError

_base_regexp = r'[a-z][a-z0-9]*(_[a-z][a-z0-9]*)*'

# attributes that trigger instantiation of user classes


supported_class_attrs = {
    'key_source', 'describe', 'alter', 'heading', 'populate', 'progress', 'primary_key',
    'proj', 'aggr', 'join', 'fetch', 'fetch1', 'head', 'tail',
    'insert1p', 'with_parts',
    'descendants', 'ancestors', 'parts', 'parents', 'children',
    'insert', 'insert1', 'update1', 'drop', 'drop_quick', 'delete', 'delete_quick'}


class TableMeta(type):
    """
    TableMeta subclasses allow applying some instance methods and properties directly
    at class level. For example, this allows Table.fetch() instead of Table().fetch().
    """
    def __getattribute__(cls, name):
        # trigger instantiation for supported class attrs
        return (cls().__getattribute__(name) if name in supported_class_attrs
                else super().__getattribute__(name))

    def __and__(cls, arg):
        return cls() & arg

    def __xor__(cls, arg):
        return cls() ^ arg

    def __sub__(cls, arg):
        return cls() - arg

    def __neg__(cls):
        return -cls()

    def __mul__(cls, arg):
        return cls() * arg

    def __matmul__(cls, arg):
        return cls() @ arg

    def __add__(cls, arg):
        return cls() + arg

    def __iter__(cls):
        return iter(cls())


class UserTable(Table, metaclass=TableMeta):
    """
    A subclass of UserTable is a dedicated class interfacing a base relation.
    UserTable is initialized by the decorator generated by schema().
    """
    # set by @schema
    _connection = None
    _heading = None
    _support = None

    # set by subclass
    tier_regexp = None
    _prefix = None

    @property
    def definition(self):
        """
        :return: a string containing the table definition using the DataJoint DDL.
        """
        raise NotImplementedError(
            'Subclasses of Table must implement the property "definition"')

    @ClassProperty
    def name(cls):
        return cls.__name__

    @ClassProperty
    def connection(cls):
        return cls._connection

    @ClassProperty
    def table_name(cls):
        """
        :return: the table name of the table formatted for mysql.
        """
        if cls._prefix is None:
            raise AttributeError('Class prefix is not defined!')
        return cls._prefix + from_camel_case(cls.name)

    @ClassProperty
    def full_table_name(cls):
        if cls not in {
            dj.Manual, dj.Imported, dj.Lookup, dj.Computed, dj.Part,
            dj.AutoImported, dj.AutoComputed, dj.Settingstable,
            UserTable
        }:
            # for derived classes only
            if cls.database is None:
                raise DataJointError((
                    'Class {0} is not properly declared '
                    '(schema decorator not applied?)'
                ).format(cls.name))
            return r"`{0:s}`.`{1:s}`".format(cls.database, cls.table_name)


class Manual(UserTable):
    """
    Inherit from this class if the table's values are entered manually.
    """
    _prefix = r''
    tier_regexp = r'(?P<manual>' + _prefix + _base_regexp + ')'


class Lookup(UserTable):
    """
    Inherit from this class if the table's values are for lookup. This is
    currently equivalent to defining the table as Manual and serves semantic
    purposes only.
    """
    _prefix = '#'
    tier_regexp = r'(?P<lookup>' + _prefix + _base_regexp.replace('TIER', 'lookup') + ')'


class Imported(UserTable, AutoPopulate):
    """
    Inherit from this class if the table's values are imported from external data sources.
    The inherited class must at least provide the function `_make_tuples`.
    """
    _prefix = '_'
    tier_regexp = r'(?P<imported>' + _prefix + _base_regexp + ')'


class Computed(UserTable, AutoPopulate):
    """
    Inherit from this class if the table's values are computed from other relations in the schema.
    The inherited class must at least provide the function `_make_tuples`.
    """
    _prefix = '__'
    tier_regexp = r'(?P<computed>' + _prefix + _base_regexp + ')'



"""
DataJoint for Python is a framework for building data piplines using MySQL databases
to represent pipeline structure and bulk storage systems for large objects.
DataJoint is built on the foundation of the relational data model and prescribes a
consistent method for organizing, populating, and querying data.

The DataJoint data model is described in https://arxiv.org/abs/1807.11104

DataJoint is free software under the LGPL License. In addition, we request
that any use of DataJoint leading to a publication be acknowledged in the publication.

Please cite:
    http://biorxiv.org/content/early/2015/11/14/031658
    http://dx.doi.org/10.1101/031658
"""

__author__ = "DataJoint Contributors"
__date__ = "November 7, 2020"
__all__ = ['__author__', '__version__',
           'config', 'conn', 'Connection',
           'Schema', 'schema', 'VirtualModule', 'create_virtual_module',
           'list_schemas', 'Table', 'FreeTable',
           'Manual', 'Lookup', 'Imported', 'Computed', 'Part',
           'AutoComputed', 'AutoImported',
           'Settingstable', 'get_settings_table',
           'Not', 'AndList', 'U', 'Diagram', 'Di', 'ERD',
           'set_password', 'kill',
           'MatCell', 'MatStruct', 'AttributeAdapter',
           'errors', 'DataJointError', 'key', 'key_hash']

from .version import __version__
from .settings import config
from .connection import conn, Connection
from .schemas import Schema
from .schemas import VirtualModule, list_schemas
from .table import Table, FreeTable
from .user_tables import Manual, Lookup, Imported, Computed
from .part_table import Part
from .auto_tables import AutoComputed, AutoImported
from .settings_table import Settingstable, get_settings_table
from .expression import Not, AndList, U
from .diagram import Diagram
from .admin import set_password, kill
from .blob import MatCell, MatStruct
from .fetch import key
from .hash import key_hash
from .attribute_adapter import AttributeAdapter
from . import errors
from .errors import DataJointError
from .migrate import migrate_dj011_external_blob_storage_to_dj012

ERD = Di = Diagram                      # Aliases for Diagram
schema = Schema                         # Aliases for Schema
create_virtual_module = VirtualModule   # Aliases for VirtualModule



import collections
import itertools
import inspect
import platform
import warnings
import numpy as np
import pandas
import logging
import uuid
import re
import warnings
from pathlib import Path
from .settings import config
from .declare import declare, alter
from .condition import make_condition
from .expression import QueryExpression
from . import blob
from .utils import user_choice, to_camel_case
from .heading import Heading
from .errors import (DuplicateError, AccessError, DataJointError, UnknownAttributeError,
                     IntegrityError)
from .version import __version__ as version

logger = logging.getLogger(__name__)

foreign_key_error_regexp = re.compile(
    r"[\w\s:]*\((?P<child>`[^`]+`.`[^`]+`), "
    r"CONSTRAINT (?P<name>`[^`]+`) "
    r"(FOREIGN KEY \((?P<fk_attrs>[^)]+)\) "
    r"REFERENCES (?P<parent>`[^`]+`(\.`[^`]+`)?) \((?P<pk_attrs>[^)]+)\)[\s\w]+\))?")

constraint_info_query = ' '.join("""
    SELECT
        COLUMN_NAME as fk_attrs,
        CONCAT('`', REFERENCED_TABLE_SCHEMA, '`.`', REFERENCED_TABLE_NAME, '`') as parent,
        REFERENCED_COLUMN_NAME as pk_attrs
    FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
    WHERE
        CONSTRAINT_NAME = %s AND TABLE_SCHEMA = %s AND TABLE_NAME = %s;
    """.split())


class _RenameMap(tuple):
    """ for internal use """
    pass


class Table(QueryExpression):
    """
    Table is an abstract class that represents a base relation, i.e. a table in the schema.
    To make it a concrete class, override the abstract properties specifying the connection,
    table name, database, and definition.
    A Relation implements insert and delete methods in addition to inherited relational operators.
    """

    _table_name = None  # must be defined in subclass
    _log_ = None  # placeholder for the Log table object

    # These properties must be set by the schema decorator (schemas.py) at class level or by FreeTable at instance level
    database = None
    declaration_context = None

    @property
    def table_name(self):
        return self._table_name

    @property
    def definition(self):
        raise NotImplementedError('Subclasses of Table must implement the `definition` property')

    def declare(self, context=None):
        """
        Declare the table in the schema based on self.definition.
        :param context: the context for foreign key resolution. If None, foreign keys are
            not allowed.
        """
        if self.connection.in_transaction:
            raise DataJointError('Cannot declare new tables inside a transaction, '
                                 'e.g. from inside a populate/make call')
        sql, external_stores = declare(self.full_table_name, self.definition, context)
        sql = sql.format(database=self.database)
        try:
            # declare all external tables before declaring main table
            for store in external_stores:
                self.connection.schemas[self.database].external[store]
            self.connection.query(sql)
        except AccessError:
            # skip if no create privilege
            pass
        else:
            self._log('Declared ' + self.full_table_name)

    def alter(self, prompt=True, context=None):
        """
        Alter the table definition from self.definition
        """
        if self.connection.in_transaction:
            raise DataJointError(
                'Cannot update table declaration inside a transaction, e.g. from inside a populate/make call')
        if context is None:
            frame = inspect.currentframe().f_back
            context = dict(frame.f_globals, **frame.f_locals)
            del frame
        old_definition = self.describe(context=context, printout=False)
        sql, external_stores = alter(self.definition, old_definition, context)
        if not sql:
            if prompt:
                print('Nothing to alter.')
        else:
            sql = "ALTER TABLE {tab}\n\t".format(tab=self.full_table_name) + ",\n\t".join(sql)
            if not prompt or user_choice(sql + '\n\nExecute?') == 'yes':
                try:
                    # declare all external tables before declaring main table
                    for store in external_stores:
                        self.connection.schemas[self.database].external[store]
                    self.connection.query(sql)
                except AccessError:
                    # skip if no create privilege
                    pass
                else:
                    self.__class__._heading = Heading(table_info=self.heading.table_info)  # reset heading
                    if prompt:
                        print('Table altered')
                    self._log('Altered ' + self.full_table_name)

    def from_clause(self):
        """
        :return: the FROM clause of SQL SELECT statements.
        """
        return self.full_table_name

    def get_select_fields(self, select_fields=None):
        """
        :return: the selected attributes from the SQL SELECT statement.
        """
        return '*' if select_fields is None else self.heading.project(select_fields).as_sql

    def parents(self, primary=None, as_objects=False, foreign_key_info=False):
        """
        :param primary: if None, then all parents are returned. If True, then only foreign keys composed of
            primary key attributes are considered.  If False, return foreign keys including at least one
            secondary attribute.
        :param as_objects: if False, return table names. If True, return table objects.
        :param foreign_key_info: if True, each element in result also includes foreign key info.
        :return: list of parents as table names or table objects
            with (optional) foreign key information.
        """
        get_edge = self.connection.dependencies.parents
        nodes = [next(iter(get_edge(name).items())) if name.isdigit() else (name, props)
                 for name, props in get_edge(self.full_table_name, primary).items()]
        if as_objects:
            nodes = [(FreeTable(self.connection, name), props) for name, props in nodes]
        if not foreign_key_info:
            nodes = [name for name, props in nodes]
        return nodes

    @property
    def has_parts(self):
        """True if self has part tables."""
        return bool(self.parts())

    @property
    def has_children(self):
        """True if self has children."""
        return bool(self.children())

    @property
    def has_parents(self):
        """True if self has parents."""
        return bool(self.parents())

    @property
    def with_parts(self):
        """Joined master with part tables. Requires that no dependent attributes match."""
        table = self
        for part_table in self.parts(as_objects=True):
            table = table * part_table
        return table

    def children(self, primary=None, as_objects=False, foreign_key_info=False):
        """
        :param primary: if None, then all children are returned. If True, then only foreign keys composed of
            primary key attributes are considered.  If False, return foreign keys including at least one
            secondary attribute.
        :param as_objects: if False, return table names. If True, return table objects.
        :param foreign_key_info: if True, each element in result also includes foreign key info.
        :return: list of children as table names or table objects
            with (optional) foreign key information.
        """
        get_edge = self.connection.dependencies.children
        nodes = [next(iter(get_edge(name).items())) if name.isdigit() else (name, props)
                 for name, props in get_edge(self.full_table_name, primary).items()]
        if as_objects:
            nodes = [(FreeTable(self.connection, name), props) for name, props in nodes]
        if not foreign_key_info:
            nodes = [name for name, props in nodes]
        return nodes

    def descendants(self, as_objects=False):
        """
        :param as_objects: False - a list of table names; True - a list of table objects.
        :return: list of tables descendants in topological order.
        """
        return [FreeTable(self.connection, node) if as_objects else node
                for node in self.connection.dependencies.descendants(self.full_table_name) if not node.isdigit()]

    def ancestors(self, as_objects=False):
        """
        :param as_objects: False - a list of table names; True - a list of table objects.
        :return: list of tables ancestors in topological order.
        """
        return [FreeTable(self.connection, node) if as_objects else node
                for node in self.connection.dependencies.ancestors(self.full_table_name) if not node.isdigit()]

    def parts(self, as_objects=False):
        """
        return part tables either as entries in a dict with foreign key informaiton or a list of objects
        :param as_objects: if False (default), the output is a dict describing the foreign keys. If True, return table objects.
        """
        self.connection.dependencies.load(force=False)
        nodes = [node for node in self.connection.dependencies.nodes
                 if not node.isdigit() and node.startswith(self.full_table_name[:-1] + '__')]
        return [FreeTable(self.connection, c) for c in nodes] if as_objects else nodes

    @property
    def is_declared(self):
        """
        :return: True is the table is declared in the schema.
        """
        return self.connection.query(
            'SHOW TABLES in `{database}` LIKE "{table_name}"'.format(
                database=self.database, table_name=self.table_name)).rowcount > 0

    @property
    def full_table_name(self):
        """
        :return: full table name in the schema
        """
        return r"`{0:s}`.`{1:s}`".format(self.database, self.table_name)

    @property
    def _log(self):
        if self._log_ is None:
            self._log_ = Log(self.connection, database=self.database, skip_logging=self.table_name.startswith('~'))
        return self._log_

    @property
    def external(self):
        return self.connection.schemas[self.database].external

    def insert1p(self, row, raise_part_missing=True, **kwargs):
        """
        Insert data records or Mapping for one entry and its part tables.
        Row must contain keys corresponding to the part table names.
        The values for the part table keys must of type dict, list of dicts,
        numpy.recarray, or pandas.DataFrame. If the value is None the part
        table will be skipped. They do not have to contain the primary keys
        of the master table.
        Does not work, if part tables have renamed primary keys.

        :param row: a pandas Series or dict-like object
        """

        if isinstance(row, pandas.DataFrame):
            # assumes all data in one dataframe
            columns = set(self.heading) & set(row)
            assert not (set(self.primary_key) - columns), \
                'not all master primary key in dataframe for insert1p.'
            # select self columns
            new_row = row[list(columns)]
            # test if self primary is unique
            assert (
                len(new_row[self.primary_key].drop_duplicates()) == 1
            ), ('master primary keys are not unique.')
            # convert master input to dictionary
            new_row = new_row.iloc[0].dropna().to_dict()

            for part_table in self.parts(as_objects=True):
                part_columns = set(part_table.heading) & set(row)
                if not part_columns:
                    new_row[part_table.name] = None
                else:
                    # select correct columns and drop duplicates
                    # TODO handling renaming
                    part_rows = row[list(part_columns)].drop_duplicates(
                        part_table.primary_key
                    )
                    new_row[part_table.name] = part_rows
            # reassign row
            row = new_row

        if not isinstance(row, pandas.Series):
            row = pandas.Series(row)

        index = set(self.heading) & set(row.index)
        # select self columns and convert to dictionary
        master_input = row[list(index)].dropna().to_dict()

        def helper():
            # insert into self
            self.insert1(master_input, **kwargs)

            # other indices
            other = set(row.index) - index
            part_table_names = []
            parts = self.parts(as_objects=True)
            # insert into part tables
            for part_table in parts:
                # if part_table exists insert otherwise skip part_table
                part_table_name = (
                    to_camel_case(part_table.table_name.split('__')[-1])
                    if not hasattr(part_table, 'name')
                    else part_table.name
                )
                if part_table_name not in row.index:
                    if raise_part_missing:
                        raise DataJointError(
                            'part table name {0} missing in index {1}'.format(
                                part_table_name, list(row.index)
                            )
                        )
                    continue
                part_table_names.append(part_table_name)

            # assert subset
            if not other.issubset(set(part_table_names)):
                raise DataJointError(
                    "Some keys could not be assigned for `inset1p`: "
                    "{0}.".format(other)
                )

            # insert into part tables
            for part_table_name, part_table in zip(part_table_names, parts):
                # try to convert to pandas Dataframe
                part_rows = row[part_table_name]
                if part_rows is None:
                    continue
                if isinstance(part_rows, dict):
                    part_rows = pandas.DataFrame([row[part_table_name]])
                else:
                    part_rows = pandas.DataFrame(row[part_table_name])
                if len(part_rows) == 0:
                    continue

                for key in self.primary_key:
                    # TODO dealing with renamed primary keys or skipped ones
                    part_rows[key] = row[key]
                # insert into part table
                part_table.insert(part_rows.to_records(False), **kwargs)

        # if in transaction skip context
        if self.connection.in_transaction:
            helper()
        else:
            with self.connection.transaction:
                helper()

    def insertp(self, rows, raise_part_missing=True, **kwargs):
        """
        Insert data records or Mapping for multiple entries and
        its part tables.
        For each row must contain keys corresponding to the part table names.
        The values for the part table keys must of type dict, numpy.recarray,
        or pandas.DataFrame. They do not have to contain the primary keys
        of the master table

        :param rows: a pandas DataFrame or numpy.recarray
        """

        rows = pandas.DataFrame(rows)

        for index, row in rows.iterrows():
            self.insert1p(
                row, raise_part_missing=raise_part_missing, **kwargs
            )

    def update1(self, row):
        """
        update1 updates one existing entry in the table.
        Caution: Updates are not part of the DataJoint data manipulation model. For strict data integrity,
        use delete and insert.
        :param row: a dict containing the primary key and the attributes to update.
        Setting an attribute value to None will reset it to the default value (if any)
        The primary key attributes must always be provided.
        Examples:
        >>> table.update1({'id': 1, 'value': 3})  # update value in record with id=1
        >>> table.update1({'id': 1, 'value': None})  # reset value to default
        """
        # argument validations
        if not isinstance(row, collections.abc.Mapping):
            raise DataJointError('The argument of update1 must be dict-like.')
        if not set(row).issuperset(self.primary_key):
            raise DataJointError('The argument of update1 must supply all primary key values.')
        try:
            raise DataJointError('Attribute `%s` not found.' % next(k for k in row if k not in self.heading.names))
        except StopIteration:
            pass  # ok
        if len(self.restriction):
            raise DataJointError('Update cannot be applied to a restricted table.')
        key = {k: row[k] for k in self.primary_key}
        if len(self & key) != 1:
            raise DataJointError('Update entry must exist.')
        # UPDATE query
        row = [self.__make_placeholder(k, v) for k, v in row.items() if k not in self.primary_key]
        query = "UPDATE {table} SET {assignments} WHERE {where}".format(
            table=self.full_table_name,
            assignments=",".join('`%s`=%s' % r[:2] for r in row),
            where=make_condition(self, key, set()))
        self.connection.query(query, args=list(r[2] for r in row if r[2] is not None))

    def insert1(self, row, **kwargs):
        """
        Insert one data record or one Mapping (like a dict).
        :param row: a numpy record, a dict-like object, or an ordered sequence to be inserted as one row.
        For kwargs, see insert()
        """
        self.insert((row,), **kwargs)

    def insert(self, rows, replace=False, skip_duplicates=False, ignore_extra_fields=False, allow_direct_insert=None):
        """
        Insert a collection of rows.
        :param rows: An iterable where an element is a numpy record, a dict-like object, a pandas.DataFrame, a sequence,
            or a query expression with the same heading as table self.
        :param replace: If True, replaces the existing tuple.
        :param skip_duplicates: If True, silently skip duplicate inserts.
        :param ignore_extra_fields: If False, fields that are not in the heading raise error.
        :param allow_direct_insert: applies only in auto-populated tables.
                                    If False (default), insert are allowed only from inside the make callback.
        Example::
        >>> relation.insert([
        >>>     dict(subject_id=7, species="mouse", date_of_birth="2014-09-01"),
        >>>     dict(subject_id=8, species="mouse", date_of_birth="2014-09-02")])
        """
        if isinstance(rows, pandas.DataFrame):
            # drop 'extra' synthetic index for 1-field index case -
            # frames with more advanced indices should be prepared by user.
            rows = rows.reset_index(
                drop=len(rows.index.names) == 1 and not rows.index.names[0]
            ).to_records(index=False)

        # prohibit direct inserts into auto-populated tables
        if not allow_direct_insert and not getattr(self, '_allow_insert', True):  # allow_insert is only used in AutoPopulate
            raise DataJointError(
                'Inserts into an auto-populated table can only done inside its make method during a populate call.'
                ' To override, set keyword argument allow_direct_insert=True.')

        if inspect.isclass(rows) and issubclass(rows, QueryExpression):   # instantiate if a class
            rows = rows()
        if isinstance(rows, QueryExpression):
            # insert from select
            if not ignore_extra_fields:
                try:
                    raise DataJointError(
                        "Attribute %s not found. To ignore extra attributes in insert, set ignore_extra_fields=True." %
                        next(name for name in rows.heading if name not in self.heading))
                except StopIteration:
                    pass
            fields = list(name for name in rows.heading if name in self.heading)
            query = '{command} INTO {table} ({fields}) {select}{duplicate}'.format(
                command='REPLACE' if replace else 'INSERT',
                fields='`' + '`,`'.join(fields) + '`',
                table=self.full_table_name,
                select=rows.make_sql(fields),
                duplicate=(' ON DUPLICATE KEY UPDATE `{pk}`={table}.`{pk}`'.format(
                    table=self.full_table_name, pk=self.primary_key[0])
                           if skip_duplicates else ''))
            self.connection.query(query)
            return

        field_list = []  # collects the field list from first row (passed by reference)
        rows = list(self.__make_row_to_insert(row, field_list, ignore_extra_fields) for row in rows)
        if rows:
            try:
                query = "{command} INTO {destination}(`{fields}`) VALUES {placeholders}{duplicate}".format(
                    command='REPLACE' if replace else 'INSERT',
                    destination=self.from_clause(),
                    fields='`,`'.join(field_list),
                    placeholders=','.join('(' + ','.join(row['placeholders']) + ')' for row in rows),
                    duplicate=(' ON DUPLICATE KEY UPDATE `{pk}`=`{pk}`'.format(pk=self.primary_key[0])
                               if skip_duplicates else ''))
                self.connection.query(query, args=list(
                    itertools.chain.from_iterable(
                        (v for v in r['values'] if v is not None) for r in rows)))
            except UnknownAttributeError as err:
                raise err.suggest(
                    'To ignore extra fields in insert, set ignore_extra_fields=True')
            except DuplicateError as err:
                raise err.suggest(
                    'To ignore duplicate entries in insert, set skip_duplicates=True')

    def delete_quick(self, get_count=False):
        """
        Deletes the table without cascading and without user prompt.
        If this table has populated dependent tables, this will fail.
        """
        query = 'DELETE FROM ' + self.full_table_name + self.where_clause()
        self.connection.query(query)
        count = self.connection.query("SELECT ROW_COUNT()").fetchone()[0] if get_count else None
        self._log(query[:255])
        return count

    def _delete_cascade(self, return_message=False):
        """service function to perform cascading deletes recursively."""
        total_count = 0
        message = ""

        # if empty nothing to delete
        if not self:
            if return_message:
                return total_count, message
            else:
                return total_count

        try:
            delete_count = self.delete_quick(get_count=True)
        except IntegrityError:
            for child, fk_info in self.children(as_objects=True, foreign_key_info=True):
                # if aliased keys rename child table
                if fk_info['aliased']:
                    child &= self.proj(**fk_info['attr_map'])
                # restriction attributes are a subset of child primary keys
                elif set(self.restriction_attributes) <= set(child.primary_key):
                    child._restriction = self._restriction
                else:
                    child &= self.proj()
                if return_message:
                    total_count_, message_ = child._delete_cascade(return_message)
                    total_count += total_count_
                    message += message_
                else:
                    total_count += child._delete_cascade(return_message)
            delete_count = self.delete_quick(get_count=True)

        total_count += delete_count
        message_ = "Deleting {count} rows from {table}\n".format(
            count=delete_count, table=self.full_table_name
        )

        if return_message:
            message += message_
            return total_count, message
        else:
            print(message_)
            return total_count

    # buggy cascade delete
    def _delete_cascade__(self):
        """service function to perform cascading deletes recursively."""
        max_attempts = 50
        delete_count = 0
        for _ in range(max_attempts):
            try:
                delete_count += self.delete_quick(get_count=True)
            except IntegrityError as error:
                match = foreign_key_error_regexp.match(error.args[0]).groupdict()
                if "`.`" not in match['child']:  # if schema name missing, use self
                    match['child'] = '{}.{}'.format(self.full_table_name.split(".")[0],
                                                    match['child'])
                if match['pk_attrs'] is not None:  # fully matched, adjusting the keys
                    match['fk_attrs'] = [k.strip('`') for k in match['fk_attrs'].split(',')]
                    match['pk_attrs'] = [k.strip('`') for k in match['pk_attrs'].split(',')]
                else:  # only partially matched, querying with constraint to determine keys
                    match['fk_attrs'], match['parent'], match['pk_attrs'] = list(map(
                        list, zip(*self.connection.query(constraint_info_query, args=(
                            match['name'].strip('`'),
                            *[_.strip('`') for _ in match['child'].split('`.`')]
                            )).fetchall())))
                    match['parent'] = match['parent'][0]
                # restrict child by self if
                # 1. if self's restriction attributes are not in child's primary key
                # 2. if child renames any attributes
                # otherwise restrict child by self's restriction.
                child = FreeTable(self.connection, match['child'])
                if set(self.restriction_attributes) <= set(child.primary_key) and \
                        match['fk_attrs'] == match['pk_attrs']:
                    child._restriction = self._restriction
                elif match['fk_attrs'] != match['pk_attrs']:
                    child &= self.proj(**dict(zip(match['fk_attrs'],
                                                  match['pk_attrs'])))
                else:
                    child &= self.proj()
                delete_count += child._delete_cascade()
            else:
                print("Deleting {count} rows from {table}".format(
                    count=delete_count, table=self.full_table_name))
                break
        else:
            raise DataJointError('Exceeded maximum number of delete attempts.')
        return delete_count

    def delete(self, transaction=True, safemode=None):
        """
        Deletes the contents of the table and its dependent tables, recursively.

        :param transaction: if True, use the entire delete becomes an atomic transaction.
        :param safemode: If True, prohibit nested transactions and prompt to confirm. Default
            is dj.config['safemode'].
        """
        safemode = config['safemode'] if safemode is None else safemode

        if config.get('delete_permission', None) is not None:
            delete_permission = config['delete_permission'](self)
            if not delete_permission:
                raise DataJointError(
                    "Cannot delete due to permission restrictions on entries."
                )

        # Start transaction
        if transaction:
            if not self.connection.in_transaction:
                self.connection.start_transaction()
            else:
                if not safemode:
                    transaction = False
                else:
                    raise DataJointError(
                        "Delete cannot use a transaction within an ongoing transaction. "
                        "Set transaction=False or safemode=False).")

        # Cascading delete
        try:
            delete_count = self._delete_cascade()
        except:
            if transaction:
                self.connection.cancel_transaction()
            raise

        # Confirm and commit
        if delete_count == 0:
            if safemode:
                print('Nothing to delete.')
            if transaction:
                self.connection.cancel_transaction()
        else:
            if not safemode or user_choice("Commit deletes?", default='no') == 'yes':
                if transaction:
                    self.connection.commit_transaction()
                if safemode:
                    print('Deletes committed.')
            else:
                if transaction:
                    self.connection.cancel_transaction()
                if safemode:
                    print('Deletes cancelled')

    def drop_quick(self):
        """
        Drops the table associated with this relation without cascading and without user prompt.
        If the table has any dependent table(s), this call will fail with an error.
        """
        if self.is_declared:
            query = 'DROP TABLE %s' % self.full_table_name
            self.connection.query(query)
            logger.info("Dropped table %s" % self.full_table_name)
            self._log(query[:255])
        else:
            logger.info("Nothing to drop: table %s is not declared" % self.full_table_name)

    def drop(self, force=False):
        """
        Drop the table and all tables that reference it, recursively.
        User is prompted for confirmation if config['safemode'] is set to True.
        """
        if self.restriction:
            raise DataJointError('A relation with an applied restriction condition cannot be dropped.'
                                 ' Call drop() on the unrestricted Table.')
        self.connection.dependencies.load()
        do_drop = True
        tables = [table for table in self.connection.dependencies.descendants(self.full_table_name)
                  if not table.isdigit()]
        if config['safemode'] and not force:
            for table in tables:
                print(table, '(%d tuples)' % len(FreeTable(self.connection, table)))
            do_drop = user_choice("Proceed?", default='no') == 'yes'
        if do_drop:
            for table in reversed(tables):
                FreeTable(self.connection, table).drop_quick()
            print('Tables dropped.  Restart kernel.')

    @property
    def size_on_disk(self):
        """
        :return: size of data and indices in bytes on the storage device
        """
        ret = self.connection.query(
            'SHOW TABLE STATUS FROM `{database}` WHERE NAME="{table}"'.format(
                database=self.database, table=self.table_name), as_dict=True).fetchone()
        return ret['Data_length'] + ret['Index_length']

    def show_definition(self):
        raise AttributeError('show_definition is deprecated. Use the describe method instead.')

    def describe(self, context=None, printout=True):
        """
        :return:  the definition string for the relation using DataJoint DDL.
        """
        if context is None:
            frame = inspect.currentframe().f_back
            context = dict(frame.f_globals, **frame.f_locals)
            del frame
        if self.full_table_name not in self.connection.dependencies:
            self.connection.dependencies.load()
        parents = self.parents(foreign_key_info=True)
        in_key = True
        definition = ('# ' + self.heading.table_status['comment'] + '\n'
                      if self.heading.table_status['comment'] else '')
        attributes_thus_far = set()
        attributes_declared = set()
        indexes = self.heading.indexes.copy()
        for attr in self.heading.attributes.values():
            if in_key and not attr.in_key:
                definition += '---\n'
                in_key = False
            attributes_thus_far.add(attr.name)
            do_include = True
            for parent_name, fk_props in parents:
                if attr.name in fk_props['attr_map']:
                    do_include = False
                    if attributes_thus_far.issuperset(fk_props['attr_map']):
                        # foreign key properties
                        try:
                            index_props = indexes.pop(tuple(fk_props['attr_map']))
                        except KeyError:
                            index_props = ''
                        else:
                            index_props = [k for k, v in index_props.items() if v]
                            index_props = ' [{}]'.format(', '.join(index_props)) if index_props else ''

                        if not fk_props['aliased']:
                            # simple foreign key
                            definition += '->{props} {class_name}\n'.format(
                                props=index_props,
                                class_name=lookup_class_name(parent_name, context) or parent_name)
                        else:
                            # projected foreign key
                            definition += '->{props} {class_name}.proj({proj_list})\n'.format(
                                props=index_props,
                                class_name=lookup_class_name(parent_name, context) or parent_name,
                                proj_list=','.join(
                                    '{}="{}"'.format(attr, ref)
                                    for attr, ref in fk_props['attr_map'].items() if ref != attr))
                            attributes_declared.update(fk_props['attr_map'])
            if do_include:
                attributes_declared.add(attr.name)
                definition += '%-20s : %-28s %s\n' % (
                    attr.name if attr.default is None else '%s=%s' % (attr.name, attr.default),
                    '%s%s' % (attr.type, ' auto_increment' if attr.autoincrement else ''),
                    '# ' + attr.comment if attr.comment else '')
        # add remaining indexes
        for k, v in indexes.items():
            definition += '{unique}INDEX ({attrs})\n'.format(
                unique='UNIQUE ' if v['unique'] else '',
                attrs=', '.join(k))
        if printout:
            print(definition)
        return definition

    def _update(self, attrname, value=None):
        """
            This is a deprecated function to be removed in datajoint 0.14. Use .update1 instead.

            Updates a field in an existing tuple. This is not a datajoyous operation and should not be used
            routinely. Relational database maintain referential integrity on the level of a tuple. Therefore,
            the UPDATE operator can violate referential integrity. The datajoyous way to update information is
            to delete the entire tuple and insert the entire update tuple.

            Safety constraints:
               1. self must be restricted to exactly one tuple
               2. the update attribute must not be in primary key

            Example:
            >>> (v2p.Mice() & key)._update('mouse_dob', '2011-01-01')
            >>> (v2p.Mice() & key)._update( 'lens')   # set the value to NULL
        """
        warnings.warn(
            '`_update` is a deprecated function to be removed in datajoint 0.14. '
            'Use `.update1` instead.')
        if len(self) != 1:
            raise DataJointError('Update is only allowed on one tuple at a time')
        if attrname not in self.heading:
            raise DataJointError('Invalid attribute name')
        if attrname in self.heading.primary_key:
            raise DataJointError('Cannot update a key value.')

        attrname, placeholder, value = self.__make_placeholder(attrname, value, False)

        command = "UPDATE {full_table_name} SET `{attrname}`={placeholder} {where_clause}".format(
            full_table_name=self.from_clause(),
            attrname=attrname,
            placeholder=placeholder,
            where_clause=self.where_clause())
        self.connection.query(command, args=(value, ) if value is not None else ())

    def _check_downstream_autopopulated(self, key=None, reload=True, error='raise'):
        """
            Check if downstream autopopulated tables include entry.
        """

        if key is not None:
            self = self & key

        if reload or not self.connection.dependencies:
            self.connection.dependencies.load()

        for child_name, child in self.children(foreign_key_info=True):

            child_table = FreeTable(self.connection, child_name)
            attr_map = {
                map_to: map_from
                for map_to, map_from in child['attr_map'].items()
                if (
                    map_from in self.heading.primary_key
                    and (map_to != map_from)
                )
            }
            restricted_table = (
                child_table
                & self.proj(**attr_map)
            )
            # recursive
            restricted_table._check_downstream_autopopulated(False, error)

            is_autopopulated = child_name.split('.')[-1].strip('`').startswith(
                ('__', '_', '_#', '#_')
            )

            if is_autopopulated and len(restricted_table) > 0:
                message = (
                    "Save update failure: Entries of downstream "
                    "autopopulated tables depend on self. "
                    "Delete appropriate entries in dependent autopopulated"
                    " tables before editing entries in self."
                )
                if error == 'ignore':
                    pass
                elif error == 'warn':
                    warnings.warn(message)
                else:
                    raise DataJointError(message)
                return False

        return True

    def save_update1(self, row, reload=True, error='raise'):
        """
            Updates multiple fields in an existing tuple, but only if no
            descendant tables are autopopulated tables with that entry.

            Safety constraints:
               1. self must be restricted to exactly one tuple
               2. the update attribute must not be in primary key

        """

        # argument validations
        if not isinstance(row, collections.abc.Mapping):
            raise DataJointError('The argument of update1 must be dict-like.')
        if not set(row).issuperset(self.primary_key):
            raise DataJointError('The argument of save_update1 must supply all primary key values.')

        key = {k: v for k, v in row.items() if k in self.primary_key}

        if len(self & key) != 1:
            raise DataJointError('Update is only allowed on one entry at a time')

        truth = self._check_downstream_autopopulated(key, reload, error)

        if not truth:
            return

        return self.update1(row)

    # --- private helper functions ----
    def __make_placeholder(self, name, value, ignore_extra_fields=False):
        """
        For a given attribute `name` with `value`, return its processed value or value placeholder
        as a string to be included in the query and the value, if any, to be submitted for
        processing by mysql API.
        :param name:  name of attribute to be inserted
        :param value: value of attribute to be inserted
        """
        if ignore_extra_fields and name not in self.heading:
            return None
        attr = self.heading[name]
        if attr.adapter:
            value = attr.adapter.put(value)
        if value is None or (attr.numeric and (value == '' or np.isnan(np.float(value)))):
            # set default value
            placeholder, value = 'DEFAULT', None
        else:  # not NULL
            placeholder = '%s'
            if attr.uuid:
                if not isinstance(value, uuid.UUID):
                    try:
                        value = uuid.UUID(value)
                    except (AttributeError, ValueError):
                        raise DataJointError(
                            'badly formed UUID value {v} for attribute `{n}`'.format(v=value,
                                                                                     n=name))
                value = value.bytes
            elif attr.is_blob:
                value = blob.pack(value)
                value = self.external[attr.store].put(value).bytes if attr.is_external else value
            elif attr.is_attachment:
                attachment_path = Path(value)
                if attr.is_external:
                    # value is hash of contents
                    value = self.external[attr.store].upload_attachment(attachment_path).bytes
                else:
                    # value is filename + contents
                    value = str.encode(attachment_path.name) + b'\0' + attachment_path.read_bytes()
            elif attr.is_filepath:
                value = self.external[attr.store].upload_filepath(value).bytes
            elif attr.numeric:
                value = str(int(value) if isinstance(value, bool) else value)
        return name, placeholder, value

    def __make_row_to_insert(self, row, field_list, ignore_extra_fields):
        """
        Helper function for insert and update
        :param row:  A tuple to insert
        :return: a dict with fields 'names', 'placeholders', 'values'
        """

        def check_fields(fields):
            """
            Validates that all items in `fields` are valid attributes in the heading
            :param fields: field names of a tuple
            """
            if not field_list:
                if not ignore_extra_fields:
                    for field in fields:
                        if field not in self.heading:
                            raise KeyError(u'`{0:s}` is not in the table heading'.format(field))
            elif set(field_list) != set(fields).intersection(self.heading.names):
                raise DataJointError('Attempt to insert rows with different fields')

        if isinstance(row, np.void):  # np.array
            check_fields(row.dtype.fields)
            attributes = [self.__make_placeholder(name, row[name], ignore_extra_fields)
                          for name in self.heading if name in row.dtype.fields]
        elif isinstance(row, collections.abc.Mapping):  # dict-based
            check_fields(row)
            attributes = [self.__make_placeholder(name, row[name], ignore_extra_fields)
                          for name in self.heading if name in row]
        else:  # positional
            try:
                if len(row) != len(self.heading):
                    raise DataJointError(
                        'Invalid insert argument. Incorrect number of attributes: '
                        '{given} given; {expected} expected'.format(
                            given=len(row), expected=len(self.heading)))
            except TypeError:
                raise DataJointError('Datatype %s cannot be inserted' % type(row))
            else:
                attributes = [self.__make_placeholder(name, value, ignore_extra_fields)
                              for name, value in zip(self.heading, row)]
        if ignore_extra_fields:
            attributes = [a for a in attributes if a is not None]

        assert len(attributes), 'Empty tuple'
        row_to_insert = dict(zip(('names', 'placeholders', 'values'), zip(*attributes)))
        if not field_list:
            # first row sets the composition of the field list
            field_list.extend(row_to_insert['names'])
        else:
            #  reorder attributes in row_to_insert to match field_list
            order = list(row_to_insert['names'].index(field) for field in field_list)
            row_to_insert['names'] = list(row_to_insert['names'][i] for i in order)
            row_to_insert['placeholders'] = list(row_to_insert['placeholders'][i] for i in order)
            row_to_insert['values'] = list(row_to_insert['values'][i] for i in order)
        return row_to_insert


def lookup_class_name(name, context, depth=3):
    """
    given a table name in the form `schema_name`.`table_name`, find its class in the context.
    :param name: `schema_name`.`table_name`
    :param context: dictionary representing the namespace
    :param depth: search depth into imported modules, helps avoid infinite recursion.
    :return: class name found in the context or None if not found
    """
    # breadth-first search
    nodes = [dict(context=context, context_name='', depth=depth)]
    while nodes:
        node = nodes.pop(0)
        for member_name, member in node['context'].items():
            if not member_name.startswith('_'):  # skip IPython's implicit variables
                if inspect.isclass(member) and issubclass(member, Table):
                    if member.full_table_name == name:   # found it!
                        return '.'.join([node['context_name'],  member_name]).lstrip('.')
                    try:  # look for part tables
                        parts = member._ordered_class_members
                    except AttributeError:
                        pass  # not a UserTable -- cannot have part tables.
                    else:
                        for part in (getattr(member, p) for p in parts if p[0].isupper() and hasattr(member, p)):
                            if inspect.isclass(part) and issubclass(part, Table) and part.full_table_name == name:
                                return '.'.join([node['context_name'], member_name, part.__name__]).lstrip('.')
                elif node['depth'] > 0 and inspect.ismodule(member) and member.__name__ != 'datajoint':
                    try:
                        nodes.append(
                            dict(context=dict(inspect.getmembers(member)),
                                 context_name=node['context_name'] + '.' + member_name,
                                 depth=node['depth']-1))
                    except ImportError:
                        pass  # could not import, so do not attempt
    return None


class FreeTable(Table):
    """
    A base relation without a dedicated class. Each instance is associated with a table
    specified by full_table_name.
    :param conn:  a dj.Connection object
    :param full_table_name: in format `database`.`table_name`
    """
    def __init__(self, conn, full_table_name):
        self.database, self._table_name = (s.strip('`') for s in full_table_name.split('.'))
        self._connection = conn
        self._support = [full_table_name]
        self._heading = Heading(table_info=dict(
            conn=conn,
            database=self.database,
            table_name=self.table_name,
            context=None))

    def __repr__(self):
        return "FreeTable(`%s`.`%s`)\n" % (self.database, self._table_name) + super().__repr__()


class Log(Table):
    """
    The log table for each schema.
    Instances are callable.  Calls log the time and identifying information along with the event.
    :param skip_logging: if True, then log entry is skipped by default. See __call__
    """

    _table_name = '~log'

    def __init__(self, conn, database, skip_logging=False):
        self.database = database
        self.skip_logging = skip_logging
        self._connection = conn
        self._heading = Heading(table_info=dict(
            conn=conn,
            database=database,
            table_name=self.table_name,
            context=None
        ))
        self._support = [self.full_table_name]

        self._definition = """    # event logging table for `{database}`
        id       :int unsigned auto_increment     # event order id
        ---
        timestamp = CURRENT_TIMESTAMP : timestamp # event timestamp
        version  :varchar(12)                     # datajoint version
        user     :varchar(255)                    # user@host
        host=""  :varchar(255)                    # system hostname
        event="" :varchar(255)                    # event message
        """.format(database=database)

        super().__init__()

        if not self.is_declared:
            self.declare()
            self.connection.dependencies.clear()
        self._user = self.connection.get_user()

    @property
    def definition(self):
        return self._definition

    def __call__(self, event, skip_logging=None):
        """
        :param event: string to write into the log table
        :param skip_logging: If True then do not log. If None, then use self.skip_logging
        """
        skip_logging = self.skip_logging if skip_logging is None else skip_logging
        if not skip_logging:
            try:
                self.insert1(dict(
                    user=self._user,
                    version=version + 'py',
                    host=platform.uname().node,
                    event=event), skip_duplicates=True, ignore_extra_fields=True)
            except DataJointError:
                logger.info('could not log event in table ~log')

    def delete(self):
        """bypass interactive prompts and cascading dependencies"""
        self.delete_quick()

    def drop(self):
        """bypass interactive prompts and cascading dependencies"""
        self.drop_quick()
